import numpy as np
import math
import time
from .stepsize_adapter import HamiltonianBasedStepsizeAdapter, initialize_stepsize
from .dynamics import HamiltonianDynamics
from .util import warn_message_only


class NoUTurnSampler():

    def __init__(self, f, mass=None, warning_requested=True):
        """
        Parameters
        ----------
        f: callable
            Return the log probability and gradient evaluated at q.
        mass: None, numpy 1d array, or callable
        """
        self.f = f
        self.dynamics = HamiltonianDynamics(mass)
        self.warning_requested = warning_requested

    def generate_samples(
            self, q0, n_burnin, n_sample, dt_range=None, seed=None, n_update=0,
            adapt_stepsize=False, target_accept_prob=.9, final_adaptsize=.05):
        """
        Implements the No-U-Turn Sampler (NUTS) of Hoffman and Gelman (2011).

        Parameters:
        -----------
        dt_range: None, float, or ndarray of length 2
        adapt_stepsize: bool
            If True, the max stepsize will be adjusted to to achieve the target
            acceptance rate. Forced to be True if dt_range is None.
        """

        if seed is not None:
            np.random.seed(seed)

        q = q0
        logp, grad = self.f(q)

        if np.isscalar(dt_range):
            dt_range = np.array(2 * [dt_range])

        elif dt_range is None:
            p = self.dynamics.draw_momentum(len(q))
            logp_joint0 = - self.dynamics.compute_hamiltonian(logp, p)
            dt = initialize_stepsize(
                lambda dt: self.compute_onestep_accept_prob(dt, q, p, grad, logp_joint0)
            )
            dt_range = dt * np.array([.8, 1.0])
            adapt_stepsize = True

        max_stepsize_adapter = HamiltonianBasedStepsizeAdapter(
            init_stepsize=1., target_accept_prob=target_accept_prob,
            reference_iteration=n_burnin, adaptsize_at_reference=final_adaptsize
        )

        if n_update > 0:
            n_per_update = math.ceil((n_burnin + n_sample) / n_update)
        else:
            n_per_update = float('inf')
        samples = np.zeros((len(q), n_sample + n_burnin))
        logp_samples = np.zeros(n_sample + n_burnin)
        accept_prob = np.zeros(n_sample + n_burnin)
        max_dt = np.zeros(n_burnin)

        tic = time.time()
        use_averaged_stepsize = False
        for i in range(n_sample + n_burnin):
            dt_multiplier \
                = max_stepsize_adapter.get_current_stepsize(use_averaged_stepsize)
            dt = np.random.uniform(dt_range[0], dt_range[1])
            dt *= dt_multiplier
            q, info = self.generate_next_state(dt, q, logp, grad)
            logp, grad = info['logp'], info['grad']
            if i < n_burnin and adapt_stepsize:
                max_dt[i] = dt_range[1] * dt_multiplier
                max_stepsize_adapter.adapt_stepsize(info['ave_hamiltonian_error'])
            elif i == n_burnin - 1:
                use_averaged_stepsize = True
            samples[:, i] = q
            logp_samples[i] = logp
            if (i + 1) % n_per_update == 0:
                print('{:d} iterations have been completed.'.format(i + 1))

        toc = time.time()
        time_elapsed = toc - tic

        info = {
            'logp_samples': logp_samples,
            'accept_prob_samples': accept_prob,
            'sampling_time': time_elapsed
        }
        if adapt_stepsize:
            info['max_stepsize'] = max_dt

        return samples, info


    def compute_onestep_accept_prob(self, dt, q0, p0, grad0, logp_joint0):
        _, p, logp, _ = self.dynamics.integrate(self.f, dt, q0, p0, grad0)
        logp_joint = - self.dynamics.compute_hamiltonian(logp, p)
        accept_prob = np.exp(logp_joint - logp_joint0)
        return accept_prob

    def generate_next_state(self, dt, q, logp=None, grad=None, p=None,
                            max_height=10, hamiltonian_error_tol=100):

        n_grad_evals = 0
        if logp is None or grad is None:
            logp, grad = self.f(q)
            n_grad_evals += 1

        if p is None:
            p = self.dynamics.draw_momentum(len(q))

        logp_joint = - self.dynamics.compute_hamiltonian(logp, p)
        logp_joint_threshold = logp_joint - np.random.exponential()
            # Slicing variable in the log-scale.

        tree = _TrajectoryTree(
            self.dynamics, self.f, dt, q, p, logp, grad, logp_joint, logp_joint,
            logp_joint_threshold, hamiltonian_error_tol
        )
        directions = 2 * (np.random.rand(max_height) < 0.5) - 1
            # Pre-allocation of random directions is unnecessary, but makes the code easier to test.
        tree, final_height, last_doubling_rejected, maxed_before_u_turn \
            = self._grow_trajectory_till_u_turn(tree, directions)
        q, logp, grad = tree.sample
        n_grad_evals += tree.n_integration_step

        if self.warning_requested:
            self._issue_warnings(
                tree.instability_detected, maxed_before_u_turn, max_height
            )

        info = {
            'logp': logp,
            'grad': grad,
            'ave_accept_prob': tree.ave_accept_prob,
            'ave_hamiltonian_error': tree.ave_hamiltonian_error,
            'n_grad_evals': n_grad_evals,
            'tree_height': final_height,
            'u_turn_detected': tree.u_turn_detected,
            'instability_detected': tree.instability_detected,
            'last_doubling_rejected': last_doubling_rejected
        }

        return q, info

    def _issue_warnings(
            self, instability_detected, maxed_before_u_turn, max_height):

        if instability_detected:
            warn_message_only(
                "Numerical integration became unstable while simulating a "
                "NUTS trajectory."
            )
        if maxed_before_u_turn:
            warn_message_only(
                'The trajectory tree reached the max height of {:d} before '
                'meeting the U-turn condition.'.format(max_height)
            )
        return

    @staticmethod
    def _grow_trajectory_till_u_turn(tree, directions):

        height = 0 # Referred to as 'depth' in the original paper, but arguably the
                   # trajectory tree is built 'upward' on top of the existing ones.
        max_height = len(directions)
        trajectory_terminated = False
        while not trajectory_terminated:

            doubling_rejected \
                = tree.double_trajectory(height, directions[height])
                # No transition to the next half of trajectory takes place if the
                # termination criteria are met within the next half tree.

            height += 1
            trajectory_terminated \
                = tree.u_turn_detected or tree.instability_detected or (height >= max_height)
            maxed_before_u_turn \
                = height >= max_height and (not tree.u_turn_detected)

        return tree, height, doubling_rejected, maxed_before_u_turn


class _TrajectoryTree():
    """
    Collection of (a subset of) states along the simulated Hamiltonian dynamics
    trajcetory endowed with a binary tree structure.
    """

    def __init__(
            self, dynamics, f, dt, q, p, logp, grad, joint_logp,
            init_joint_logp, joint_logp_threshold, hamiltonian_error_tol=100.,
            u_turn_criterion='momentum'):

        self.dynamics = dynamics
        self.f = f
        self.dt = dt
        self.joint_logp_threshold = joint_logp_threshold
        self.front_state = (q, p, grad)
        self.rear_state = (q, p, grad)
        self.sample = (q, logp, grad)
        self.u_turn_detected = False
        self.min_hamiltonian = - joint_logp
        self.max_hamiltonian = - joint_logp
        self.hamiltonian_error_tol = hamiltonian_error_tol
        self.n_acceptable_state = int(joint_logp > joint_logp_threshold)
        self.n_integration_step = 0
        self.init_joint_logp = init_joint_logp
        self.height = 0
        self.ave_hamiltonian_error = abs(init_joint_logp - joint_logp)
        self.ave_accept_prob = min(1, math.exp(joint_logp - init_joint_logp))
        self.velocity_based_u_turn = (u_turn_criterion == 'velocity')

    @property
    def n_node(self):
        return 2 ** self.height

    @property
    def instability_detected(self):
        fluctuation_along_trajectory = self.max_hamiltonian - self.min_hamiltonian
        return fluctuation_along_trajectory > self.hamiltonian_error_tol

    def double_trajectory(self, height, direction):
        next_tree = self._build_next_tree(
            *self._get_states(direction), height, direction
        )
        no_transition_to_next_tree_attempted \
            = self._merge_next_tree(next_tree, direction, sampling_method='swap')
        return no_transition_to_next_tree_attempted

    def _build_next_tree(self, q, p, grad, height, direction):

        if height == 0:
            return self._build_next_singleton_tree(q, p, grad, direction)

        subtree = self._build_next_tree(q, p, grad, height - 1, direction)
        trajectory_terminated_within_subtree \
            = subtree.u_turn_detected or subtree.instability_detected
        if not trajectory_terminated_within_subtree:
            next_subtree = self._build_next_tree(
                *subtree._get_states(direction), height - 1, direction
            )
            subtree._merge_next_tree(next_subtree, direction, sampling_method='uniform')

        return subtree

    def _build_next_singleton_tree(self, q, p, grad, direction):
        q, p, logp, grad = \
            self.dynamics.integrate(self.f, direction * self.dt, q, p, grad)
        self.n_integration_step += 1
        if math.isinf(logp):
            joint_logp = - float('inf')
        else:
            joint_logp = - self.dynamics.compute_hamiltonian(logp, p)
        return self._clone_tree(q, p, logp, grad, joint_logp)

    def _clone_tree(self, q, p, logp, grad, joint_logp):
        """ Construct a tree with shared dynamics and acceptance criteria. """
        return _TrajectoryTree(
            self.dynamics, self.f, self.dt, q, p, logp, grad, joint_logp, self.init_joint_logp,
            self.joint_logp_threshold, self.hamiltonian_error_tol
        )

    def _merge_next_tree(self, next_tree, direction, sampling_method):

        # Trajectory termination flags from the next tree must be propagated up
        # the call stack, but other states of the tree is updated only if the
        # next tree is accessible from the current tree (i.e. the trajectory
        # did not get terminated within the next tree).

        self.u_turn_detected = self.u_turn_detected or next_tree.u_turn_detected
        self.min_hamiltonian = min(self.min_hamiltonian, next_tree.min_hamiltonian)
        self.max_hamiltonian = max(self.max_hamiltonian, next_tree.max_hamiltonian)
        trajectory_terminated_within_next_tree \
            = next_tree.u_turn_detected or next_tree.instability_detected

        if not trajectory_terminated_within_next_tree:
            self._update_sample(next_tree, sampling_method)
            self.n_acceptable_state += next_tree.n_acceptable_state
            self._set_states(*next_tree._get_states(direction), direction)
            self.u_turn_detected \
                = self.u_turn_detected or self._check_u_turn_at_front_and_rear_ends()
            weight = self.n_node / (self.n_node + next_tree.n_node)
            self.ave_hamiltonian_error \
                = weight * self.ave_hamiltonian_error + (1 - weight) * next_tree.ave_hamiltonian_error
            self.ave_accept_prob \
                = weight * self.ave_accept_prob + (1 - weight) * next_tree.ave_accept_prob
            self.height += 1

        return trajectory_terminated_within_next_tree

    def _update_sample(self, next_tree, method):
        """
        Parameters
        ----------
        method: {'uniform', 'swap'}
        """
        if method == 'uniform':
            n_total = self.n_acceptable_state + next_tree.n_acceptable_state
            sampling_weight_on_next_tree \
                = next_tree.n_acceptable_state / max(1, n_total)
        elif method == 'swap':
            sampling_weight_on_next_tree \
                = next_tree.n_acceptable_state / self.n_acceptable_state
        if np.random.uniform() < sampling_weight_on_next_tree:
            self.sample = next_tree.sample

    def _check_u_turn_at_front_and_rear_ends(self):
        q_front, p_front, _ = self._get_states(1)
        q_rear, p_rear, _ = self._get_states(-1)
        dq = q_front - q_rear
        if self.velocity_based_u_turn:
            v_front = self.dynamics.convert_to_velocity(p_front)
            v_rear = self.dynamics.convert_to_velocity(p_rear)
            u_turned = (np.dot(dq, v_front) < 0) or (np.dot(dq, v_rear) < 0)
        else:
            u_turned = (np.dot(dq, p_front) < 0) or (np.dot(dq, p_rear) < 0)
        return u_turned

    def _set_states(self, q, p, grad, direction):
        if direction > 0:
            self.front_state = (q, p, grad)
        else:
            self.rear_state = (q, p, grad)

    def _get_states(self, direction):
        if direction > 0:
            return self.front_state
        else:
            return self.rear_state



def dot(self, x, transpose=False):
    result = np.zeros(self.shape[transpose])
    cdef
    double[:]
    x_view = x
    cdef
    double[:]
    result_view = result
    matvec_status = mkl_plain_matvec(
        self.A, & x_view[0], & result_view[0], int(transpose)
    )
    return result



cdef mkl_plain_matvec(
		sparse_matrix_t A, const double* x, double* result, bint transpose
	):

	cdef sparse_operation_t operation
	if transpose:
		operation = SPARSE_OPERATION_TRANSPOSE
	else:
		operation = SPARSE_OPERATION_NON_TRANSPOSE
	cdef double alpha = 1.
	cdef double beta = 0.
	cdef matrix_descr mat_descript
	mat_descript.type = SPARSE_MATRIX_TYPE_GENERAL
	status = mkl_sparse_d_mv(
		operation, alpha, A, mat_descript, x, beta, result
	)
	return status


from functools import wraps
import numpy as np

@staticmethod
def precondition_logp_and_grad_by_scaling(
        compute_logp_and_grad, scale, jacobian_adjusted=True):
    """
    Decorate a function for computing the log-density and gradient so that
    it acts on the scaled parameter `param / scale`.

    Parameters
    ----------
    compute_logp_and_grad(parameter, loglik_only): callable
    scale: ndarray
    jacobian_adjusted: bool
        If `True`, includes the Jacobian factor due to the transformation
        `(param, scale) -> (param / scale, scale)`. Set to `False` if
        preconditioning a likelihood.
    """

    @wraps(compute_logp_and_grad)
    def f(param_scaled, loglik_only=False):
        param = scale * param_scaled
        logp, grad = compute_logp_and_grad(param, loglik_only)
        if not loglik_only:
            grad *= scale  # Chain rule
        if jacobian_adjusted:
            log_det = - np.sum(scale)
            logp += log_det
        return logp, grad

    return f

def adjust_for_logtransform(logp, grad, param, log_param):
    # Chain rule
    grad *= param
    # Jacobian adjustment
    logp += np.sum(log_param)
    grad += 1.
    return logp, grad


import numpy as np
import scipy as sp
import pickle as pkl
import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter

import sys
sys.path.insert(0, '..')
from bayesbridge.mcmc_diagnostics import estimate_ess


class SimulationSummarizer(object):

    def __init__(self):
        pass

    def compute_std_difference(self, samples_1, samples_2, moment=1, n_burnin=0):
        # Compute the difference in the mean of the two samples and
        # estimate the Monte Carlo error of the difference.

        mean_1, ess_1, mc_var_1 = self.estimate_monte_carlo_error(samples_1 ** moment, n_burnin)
        mean_2, ess_2, mc_var_2 = self.estimate_monte_carlo_error(samples_2 ** moment, n_burnin)
        diff = mean_1 - mean_2
        mc_var = mc_var_1 + mc_var_2
        std_diff = diff / np.sqrt(mc_var)
        return mean_1, mean_2, ess_1, ess_2, std_diff

    @staticmethod
    def estimate_monte_carlo_error(samples, n_burnin=0):
        post_burnin = samples[:, n_burnin:]
        mean = np.mean(post_burnin, 1)
        ess = estimate_ess(post_burnin, axis=-1)
        mc_var = np.var(post_burnin, axis=1) / ess
        return mean, ess, mc_var

    def plot_refline(self, x, y, label=None):
        plt.plot(
            [x.min(), x.max()],
            [y.min(), y.max()],
            '--',
            color='tab:green',
            label=label
        )

    def plot_posterior_summary_comparison(self, direct_mean, cg_mean, moment=1, label="ref line: x = y"):
        subset_ind = (direct_mean > - float('inf')) # Don't remember why I did this
        self.plot_refline(
            direct_mean[subset_ind],
            cg_mean[subset_ind],
            label=label
        )
        plt.plot(
            direct_mean[subset_ind],
            cg_mean[subset_ind],
            'o', fillstyle='none', ms=12,
            rasterized=True
        )
        legend = None
        if label is not None:
            legend = plt.legend(loc=[.26, .08], frameon=False)
        x_estimand, y_estimand = self.get_estimand_str(moment)
        plt.xlabel(r'${:s}$'.format(x_estimand))
        plt.ylabel(r'${:s}$'.format(y_estimand))
        plt.gca().set_aspect('equal', 'box')
        return legend

    def plot_std_diff_hist(self, std_diff, moment=1):

        width = max(np.abs(np.min(std_diff)), np.max(std_diff))
        bins = np.linspace(-width, width, 51)

        plt.hist(std_diff, bins=bins, density=True,
                 label='observed')
        x = np.linspace(-width, width, 101)
        plt.plot(x, sp.stats.norm.pdf(x), 'C2-',
                 label='null')
        bench_estimand, cg_estimand = self.get_estimand_str(moment, hatted=True)
        plt.xlabel(
            r'$\left( {:s} - {:s} \right) '.format(cg_estimand, bench_estimand)
            + '\, / \, \hat{\sigma}_j$'
        )
        plt.yticks([])
        legend = plt.legend(loc='upper right', frameon=False)
        # plt.xlim([-4.2, 4.2])
        return legend

    def plot_ess_comparion(self, direct_ess, cg_ess, moment=1, label=None):

        plt.loglog(
            direct_ess, cg_ess, 'o',
            alpha=.25, rasterized=True
        )
        self.plot_refline(direct_ess, cg_ess, label)

        ticks = [500, 1000, 2000]
        tick_labels = [r' $5 \! \times \! 10^2$', r'$10^3$', r' $2 \! \times \! 10^3$']
        ax = plt.gca()
        ax.xaxis.set_major_formatter(NullFormatter())
        ax.xaxis.set_minor_formatter(NullFormatter())
        ax.yaxis.set_major_formatter(NullFormatter())
        ax.yaxis.set_minor_formatter(NullFormatter())
            # Deal with bug https://github.com/matplotlib/matplotlib/issues/8027/
        plt.xticks(ticks, tick_labels)
        plt.yticks(ticks, tick_labels)

        x_estimand, y_estimand = self.get_estimand_str(moment)
        x_label = r'ESS of ${:s}$'.format(x_estimand)
        y_label = r'ESS of ${:s}$'.format(y_estimand)
        plt.xlabel(x_label)
        plt.ylabel(y_label)

        plt.gca().set_aspect('equal', 'box')

    def get_estimand_str(self, moment, hatted=False):
        if hatted:
            coef_symbol = r'\hat{\beta}'
        else:
            coef_symbol = r'\beta'
        x_estimand = coef_symbol + r'_{{\rm bench}, j}'
        y_estimand = coef_symbol + r'_{{\rm cg}, j}'
        if moment == 2:
            x_estimand = r'{:s}^{:d}'.format(x_estimand, moment)
            y_estimand = r'{:s}^{:d}'.format(y_estimand, moment)
        return x_estimand, y_estimand

    def get_linalg_cost(self, n_obs, n_pred, n_thread):

        if n_thread == 1:
            matvec_cost = [.0236, .0933, .414]
            matmat_cost = [5.57, 43.5, 345]
            chol_cost = [0.932, 6.95, 52.1]
        elif n_thread == 4:
            matvec_cost = [.0205, .0822, .334]
            matmat_cost = [1.88, 15.1, 113]
            chol_cost = [0.422, 2.87, 19.9]
        else:
            raise ValueError()

        index = {
            (12500, 5000): 0,
            (25000, 10000): 1,
            (50000, 20000): 2,
        }[(n_obs, n_pred)]

        return matvec_cost[index], matmat_cost[index], chol_cost[index]




import numpy as np

class RegressionCoeffficientPosteriorSummarizer():

    def __init__(self, n_coef, n_unshrunk, regularizing_slab_size,
                 pc_summary_method='average'):
        self.n_unshrunk = n_unshrunk
        self.beta_scaled_summarizer = OntheflySummarizer(n_coef)
        self.slab_size = regularizing_slab_size
        self.pc_summarizer = DirectionSummarizer(pc_summary_method)

    def scale_beta(self, beta, gscale, lscale):
        beta_scaled = beta.copy()
        beta_scaled[self.n_unshrunk:] \
            /= self.compute_prior_scale(gscale, lscale)
        return beta_scaled

    def update(self, beta, gscale, lscale):
        beta_scaled = self.scale_beta(beta, gscale, lscale)
        self.beta_scaled_summarizer.update_stats(beta_scaled)

    def update_precond_hessian_pc(self, pc):
        self.pc_summarizer.update(pc)

    def extrapolate_beta_condmean(self, gscale, lscale):
        beta_condmean_guess = self.beta_scaled_summarizer.stats['mean'].copy()
        beta_condmean_guess[self.n_unshrunk:] \
            *= self.compute_prior_scale(gscale, lscale)
        return beta_condmean_guess

    def estimate_beta_precond_scale_sd(self):
        return self.beta_scaled_summarizer.estimate_post_sd()

    def estimate_precond_hessian_pc(self):
        return self.pc_summarizer.get_mean()

    def compute_prior_scale(self, gscale, lscale):
        """ Compute the regularized prior scale in a numerically stable way. """
        unreg_prior_scale = gscale * lscale
        return unreg_prior_scale \
               / np.sqrt(1 + (unreg_prior_scale / self.slab_size) ** 2)

class DirectionSummarizer():

    def __init__(self, summary_method):
        """
        Parameters
        ----------
        summary_method: str, {'average', 'previous'}
        """
        self.method = summary_method
        self.n_averaged = 0
        self.v = None

    def update(self, v):
        if self.n_averaged == 0 or self.method == 'previous':
            self.v = v
        else:
            v *= np.sign(np.inner(self.v, v))
            weight = 1 / (1 + self.n_averaged)
            self.v = weight * v + (1 - weight) * self.v
        self.n_averaged += 1

    def get_mean(self):
        return self.v


class OntheflySummarizer():
    """
    Carries out online updates of the mean, variance, and other statistics of a
    random sequence.
    """

    def __init__(self, n_param, sd_prior_samplesize=5):
        """

        Params
        ------
        init: dict
        sd_prior_samplesize: int
            Weight on the initial estimate of the posterior standard
            deviation; the estimate is treated as if it is an average of
            'prior_samplesize' previous values.
        """
        self.sd_prior_samplesize = sd_prior_samplesize
        self.sd_prior_guess = np.ones(n_param)
        self.n_averaged = 0
        self.stats = {
            'mean': np.zeros(n_param),
            'square': np.ones(n_param)
        }

    def update_stats(self, theta):

        weight = 1 / (1 + self.n_averaged)
        self.stats['mean'] = (
            weight * theta + (1 - weight) * self.stats['mean']
        )
        self.stats['square'] = (
            weight * theta ** 2
            + (1 - weight) * self.stats['square']
        )
        self.n_averaged += 1

    def estimate_post_sd(self):

        # TODO: implment Welford's algorithm for better numerical accuracy.
        mean = self.stats['mean']
        sec_moment = self.stats['square']

        if self.n_averaged > 1:
            var_estimator = self.n_averaged / (self.n_averaged - 1) * (
                sec_moment - mean ** 2
            )
            estimator_weight = (self.n_averaged - 1) \
                / (self.n_averaged - 1 + self.sd_prior_samplesize)
            sd_estimator = np.sqrt(
                estimator_weight * var_estimator \
                + (1 - estimator_weight) * self.sd_prior_guess ** 2
            )
        else:
            sd_estimator = self.sd_prior_guess

        return sd_estimator


import sys

def main():
    filename = sys.argv[1] 
    newfilename = "{:s}_converted.py".format(
        filename[:filename.find('.')]
    )
    with open(filename, 'r') as file, open(newfilename, 'w') as newfile:
        for line in file:
            line = convert_basic_syntax(line)
            line = convert_to_numpy_syntax(line)
            line = convert_to_matplotlib_syntax(line)
            newfile.write(line)

def convert_basic_syntax(str):
    # Remove a line if it is a closing of a control statement. 
    if str.lstrip()[:3] == 'end':
        str = ''
        return str

    str = str.replace('%', '#')
    str = str.replace(';', '')
    str = str.replace('...', '\\')
    str = str.replace('~', '_')
    str = str.replace('length(', 'len(')
    
    # Add a colon at the end of control statements. Algorithmically
    # converting a for-loop syntax seems trickly, so not delt here.
    control = ['for', 'if', 'else', 'while']
    wo_lspace = str.lstrip()
    for word in control:
        if wo_lspace[:len(word)] == word:
            str = str.rstrip('\n') + ':' + '\n'
    
    return str

def convert_to_numpy_syntax(str):
    # For the functions with numpy equivalents, simply append 'np.'.
    # 'zeros' and 'ones' need further processing inside the parentheses.
    # Also watch out for the fact that axis indexing starts with 0 in Python.
    expression_list = [
        'zeros', 'ones', 'exp', 'log', 'sqrt', 'abs', 'sign', 'linspace'
    ]
    for expression in expression_list:
        str = str.replace(expression + '(', 'np.' + expression + '(')
        
    str = str.replace('rng(', 'np.random.seed(')
        
    return str

def convert_to_matplotlib_syntax(str):
    # Remove the lines 'hold on' and 'hold off'
    if str.lstrip()[:5] == 'hold ':
        return ''
    
    expression_list = [
        'plot', 'subplot', 'hist', 'refline'
    ]
    for expression in expression_list:
        wo_lspace = str.lstrip()
        # Avoid matching 'plot' within a string 'subplot'.
        if wo_lspace[:len(expression)] == expression:
            str = str.replace(expression + '(', 'plt.' + expression + '(')

    return str

if __name__ == "__main__":
    main()


"""
Provides some of the most commonly used estimators of the effective sample
sizes in analyzing Markov chain Monte Carlo outputs.

For theories and ideas behind each of the algorithms, see the survey paper
below and the references therein:

    M. Thompson (2010). A Comparison of Methods for Computing Autocorrelation Time.
    https://arxiv.org/abs/1011.0175
"""

import math
import warnings
from . import ar_model
import numpy as np

# Monkey patch warnings
warnings.formatwarning = lambda message, category, filename, lineno, line=None: (
    '{:s}:{:d}: {:s}: {:s}\n'.format(filename, lineno, category.__name__, str(message))
)


def estimate_ess(samples, axis=0, method='ar', normed=False, options={}):
    """
    Wrapper for calling a function specified by the `method` argument.
    Estimates effective sample sizes (ESS) of samples along the specified axis.

    Parameters
    ----------
    samples : numpy 1d or 2d array
    axis : int, {0, 1}
    method : str, {'ar', 'monotone-sequence', 'batch-means'}
    normed : bool
        If True, the efficiency (ESS per MCMC sample) is returned.
    options : dict with keys in ['max_ar_order', 'n_batch']

    Returns
    -------
    ess : numpy array
    """

    if type(samples) is not np.ndarray:
        raise TypeError("The first argument must be a numpy array.")
    if samples.ndim > 2:
        raise ValueError("The function only supports a 1d or 2d array.")
    if samples.ndim == 2 and axis == -1:
        axis = 1
    if axis not in [0, 1]:
        raise ValueError("Invalid axis value.")

    if samples.shape[axis] <= 25:
        warnings.warn(
            "The number of samples is extremely small. The estimated ESS "
            "will likely be not reliable at all."
        )

    if method == 'ar':
        max_ar_order = options.get('max_ar_order', None)
        ess = ar_process_fit(samples, axis, normed, max_ar_order=max_ar_order)

    elif method == 'monotone-sequence':
        ess, _ = monotone_sequence(samples, axis, normed)

    elif method == 'batch-means':
        n_batch = options.get('n_batch', 25)
        ess = batch_means(samples, axis, normed, n_batch=n_batch)

    else:
        raise NotImplementedError("Method {:s} is not supported.".format(method))

    return ess


def ar_process_fit(samples, axis=0, normed=False, max_ar_order=None):
    """
    Estimates effective sample sizes of samples along the specified axis by
    fitting an autoregressive process via the Yule-Walker equation. The order
    of the AR process is determined via AIC.
    """

    if samples.ndim == 1:
        samples = samples[:, np.newaxis]

    series_length = samples.shape[axis]
    if max_ar_order is None:
        if series_length <= 100:
            max_ar_order = math.ceil(series_length / 5)
        else:
            max_ar_order = math.ceil(10 * np.log10(series_length))

    n_param = samples.shape[1 - axis]

    if series_length == 1: # Edge case
        ess = np.zeros(n_param)
    else:
        if axis == 0:
            samples = samples.T
        ess = np.array([
            _ar_process_fit_1d(x, max_ar_order) for x in samples
            # Loop is over the rows of samples.
        ])

    if normed: ess /= series_length

    return ess


def _ar_process_fit_1d(x, max_ar_order):

    ar_order, ar_coef = ar_model.fit(x, max_ar_order)

    if ar_order == 0:
        auto_corr_time = 1
    else:
        x_std = (x - np.mean(x)) / np.std(x)
        acorr = np.array([
            _compute_auto_corr(x_std, lag) for lag in range(1, ar_order + 1)
        ])
        auto_corr_time = (1 - np.inner(acorr, ar_coef)) / (1 - np.sum(ar_coef)) ** 2

    ess = len(x) / auto_corr_time
    return ess


def batch_means(samples, axis=0, normed=False, n_batch=25):
    """
    Estimates effective sample sizes of samples along the specified axis
    with the method of batch means.
    """

    if samples.ndim == 1:
        samples = samples[:, np.newaxis]

    n_sample = samples.shape[axis]
    if 2 * n_batch > n_sample:
        raise ValueError(
            "The number of batches must be less than twice the number of samples."
        )

    batch_index = np.linspace(0, n_sample, n_batch + 1).astype('int')
    batch_list = [
        np.take(samples, np.arange(batch_index[i], batch_index[i + 1]), axis)
        for i in range(n_batch)
    ]
    batch_mean = np.stack((np.mean(batch, axis) for batch in batch_list), axis)
    mcmc_var = n_sample / n_batch * np.var(batch_mean, axis)
    ess = np.var(samples, axis) / mcmc_var
    if not normed: ess *= n_sample

    return ess


def monotone_sequence(
        samples, axis=0, normed=False, require_acorr=False):
    """
    Estimates effective sample sizes of samples along the specified axis
    with the monotone positive sequence estimator of "Practical Markov
    Chain Monte Carlo" by Geyer (1992). The estimator is ONLY VALID for
    reversible Markov chains. The inputs 'mu' and 'sigma_sq' are optional
    and unnecessary for the most cases in practice.

    Parameters
    ----------
    require_acorr : bool
        If true, a list of estimated auto correlation sequences are returned.

    Returns
    -------
    ess : numpy array
    auto_cor : list of numpy array
        auto-correlation estimates of the chain up to the lag beyond which the
        auto-correlation can be considered insignificant by the monotonicity
        criterion.
    """

    if samples.ndim == 1:
        samples = samples[:, np.newaxis]

    n_param = samples.shape[1 - axis]
    n_sample = samples.shape[axis]
    ess = np.zeros(n_param)
    if n_sample <= 2: # Edge case.
        return ess

    auto_cor = []
    for j in range(n_param):
        if axis == 0:
            x = samples[:, j]
        else:
            x = samples[j, :]
        x_std = (x - np.mean(x)) / np.std(x)
        ess_j, auto_cor_j = _monotone_sequence_1d(x_std, require_acorr)
        ess[j] = ess_j
        if require_acorr:
            auto_cor.append(auto_cor_j)
    if normed:
        ess /= n_sample

    return ess, auto_cor


def _monotone_sequence_1d(x, require_acorr):
    """ The time series `x` is assumed to be standardized. """

    auto_corr = []

    # lag in [0, 1] case.
    lag_one_auto_corr = _compute_auto_corr(x, lag=1)
    running_min = 1. + lag_one_auto_corr
    auto_corr_sum = 1. + 2 * lag_one_auto_corr
    if require_acorr:
        auto_corr.extend((1., lag_one_auto_corr))
    curr_lag = 2

    while curr_lag + 2 < len(x):

        even_auto_corr, odd_auto_corr = [
            _compute_auto_corr(x, lag) for lag in [curr_lag, curr_lag + 1]
        ]
        curr_lag += 2
        if even_auto_corr + odd_auto_corr < 0:
            break

        running_min = min(running_min, (even_auto_corr + odd_auto_corr))
        auto_corr_sum += 2 * running_min
        if require_acorr:
            auto_corr.extend((even_auto_corr, odd_auto_corr))

    ess = len(x) / auto_corr_sum
    if auto_corr_sum < 0:
        # Rare, but can happen with floating point errors when the time series
        # `x` shows strong negative correlations.
        ess = float('inf')

    return ess, np.array(auto_corr)


def _compute_auto_corr(x, lag):
    """
    Returns an estimate of the lag 'k' auto-correlation of a time series 'x'.
    The estimator is biased towards zero due to the factor (len(x) - lag) / len(x).
    See Geyer (1992) Section 3.1 and the reference therein for justification.
    """
    acorr = np.mean(x[:-lag] * x[lag:]) * (len(x) - lag) / len(x)
    return acorr


def _r_coda(samples, axis=0, normed=False, R_call='rpy', n_digit=18):
    """
    Estimates effective sample sizes of samples along the specified axis.

    Parameters
    ----------
    R_call : {'rpy', 'external'}
        If 'external', the samples are first saved as a csv file and the
        method runs an R script from bash.
    """
    if R_call == 'rpy':
        ess = _coda_rpy(samples, axis, normed)
    elif R_call == 'external':
        ess = _coda_external(samples, axis, normed, n_digit)
    else:
        raise NotImplementedError()
    return ess


def _coda_rpy(samples, axis=0, normed=False):
    """
    Estimates effective sample sizes of samples along the specified axis by
    calling the R package 'coda' via rpy2. Requires the package rpy2 to be installed.
    """

    try:
        import rpy2
    except ImportError:
        warnings.warn(
            'rpy2 needs to be installed to call this function.'
        )
        raise

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    import rpy2.robjects.packages as rpackages
    from rpy2.rinterface import RRuntimeError

    try:
        coda = rpackages.importr('coda')
    except RRuntimeError:
        warnings.warn('R CODA package not installed. Installing....')
        utils = rpackages.importr('utils')
        utils.install_packages('coda')

    ess = np.squeeze(np.array([
        coda.effectiveSize(np.take(samples, i, axis))
        for i in range(samples.shape[axis])
    ]))
    if normed:
        ess = ess / samples.shape[axis]

    return ess


def _coda_external(samples, axis=0, normed=False, n_digit=18,
                   saveto_fname=None, loadfrom_fname=None, rscript_name=None):
    """
    Estimates effective sample sizes of samples along the specified axis by
    calling the R package 'coda' externally. It is a hacky but convenient way
    to call an R function without having to install rpy2 and its dependencies.
    """

    import os
    import sys

    if not (sys.platform.startswith('linux') or sys.platform == 'darwin'):
        raise OSError(
            "The current implementation uses bash commnads, so runs only on "
            "Linux and Mac."
        )

    filenum = np.random.randint(2 ** 31)
        # Append a random number to a file name to avoid conflicts.
    if saveto_fname is None:
        saveto_fname = 'mchain{:d}.csv'.format(filenum)
    if loadfrom_fname is None:
        loadfrom_fname = 'ess{:d}.csv'.format(filenum)
    if rscript_name is None:
        rscript_name = "compute_ess_with_coda.R"

    if axis == 0:
        np.savetxt(saveto_fname, samples, delimiter=',', fmt='%.{:d}e'.format(n_digit))
    else:
        np.savetxt(saveto_fname, samples.T, delimiter=',', fmt='%.{:d}e'.format(n_digit))

    # Write an R script for computing ESS with the 'coda' package if the script
    # is not already present.
    r_code = "\"args <- commandArgs(trailingOnly=T) # Read in the input and output file names\n" \
             + "x <- read.csv(args[1], header=F)\n" \
             + "library(coda)\n" \
             + "ess <- unlist(lapply(x, effectiveSize))\n" \
             + "write.table(ess, args[2], sep=',', row.names=F, col.names=F)\""
    os.system(" ".join(["[[ ! -f", rscript_name, "]] && echo", r_code, ">>", rscript_name]))

    # Write the data to a text file, read into the R script, and output
    # the result back into a text file.
    exit_status = os.system(
        " ".join(["Rscript", rscript_name, saveto_fname, loadfrom_fname])
    )
    if exit_status != 0:
        raise RuntimeError("Command line call to an Rscript failed.")

    ess = np.loadtxt(loadfrom_fname, delimiter=',').copy()
    if normed:
        ess = ess / samples.shape[axis]
    os.system(" ".join(["rm -f", saveto_fname, loadfrom_fname]))
    os.system(" ".join(["rm -f", rscript_name]))

    return ess


import sys
sys.path.insert(0, '.')

import time
import numpy as np

import data_simulation_manager
from .data_simulation_manager import DataSimulationManager


class SyntheticSimulationManager(object):

    def __init__(self, save_folder='/Users/aki-nishimura/OneDrive/OHDSI/Data/anti_coagulant/'):
        self.dsm = DataSimulationManager(path_to_data=save_folder)

    def set_simulation_params(
            self, n, p, n_signal, corr_design=True, n_factor=100, seed=0
        ):
        self.n = n
        self.p
        beta_true = np.zeros(p)
        beta_true[:n_nonzero_coef] = 1
        y, X = data_simulation_manager.simulate_data(
            n, p, beta_true, link, seed, corr_design, n_factor
        )

    def run(self, n, p, n_signal, corr_design=True, n_factor=100, seed=0,
            n_full_bayes_iter=1000, n_empir_bayes_iter=5000):

        self.dsm.set_data_generation_info(
            n, p, n_signal, corr_design, n_factor=n_factor
        )
        regress_on = self.dsm.get_model_name(n_signal, corr_design, n_factor)
        print("Regressing on " + regress_on)

        # First run full Bayes to estimate global scale.
        beta_true = np.zeros(p)
        beta_true[:n_signal] = 1
        y, X = data_simulation_manager.simulate_data(
            n, p, beta_true, 'logit', seed, corr_design, n_factor
        )
        init = {
            'beta': np.concatenate(([0], beta_true)),
        }
        self.run_full_bayes_gibbs(y, X, init, n_full_bayes_iter)

        # Fix global scale, run cg and direct for the same number of iterations

        pass

    def run_full_bayes_gibbs(self, y, X, init, n_post_burnin):

        bridge = BayesBridge(y, X, model='logit', add_intercept=True)
        start_time = time.time()
        mcmc_output = bridge.gibbs(
            n_burnin, n_post_burnin, thin, reg_exponent, init, mvnorm_method,
            seed=seed, global_scale_update=global_shrinkage_update,
            params_to_save=['beta', 'global_scale', 'logp']
        )
        end_time = time.time()
        print("Posterior computation took {:3g} seconds.".format(
            end_time - start_time
        ))
        mcmc_output['_random_gen_state'] = None  # Make it picklable

        dsm.set_data_generation_info(n, p, n_nonzero_coef, corr_design,
                                     n_factor=n_factor)
        dsm.save_gibbs_output(mcmc_output)




import numpy as np
import scipy as sp
import scipy.sparse
import pandas as pd
import pickle as pkl
from warnings import warn

class DataManager(object):

    # TODO: Consider including the interaction between treatment and propensity strata.
    def __init__(
            self, path_to_data, outcome_name='cardio_disease',
            stratified=True, include_main_effect=True, include_interaction=True,
            include_gender_age_interaction=False, stratified_by_menopause=False,
            exclude_subgroup_indicator=False, n_strata=5, baseline_strata=None,
            treatment_encoding='thiazide'):

        if outcome_name not in ['cardio_disease', 'angioedema']:
            raise ValueError('Invalid outcome name.')

        if stratified_by_menopause and include_gender_age_interaction:
            warn("Gender-age interaction not included when stratified "
                 "by the menopausal states.")
            include_gender_age_interaction = False

        if not stratified:
            n_strata = 1
        if baseline_strata is None:
            baseline_strata = int(n_strata / 2)
        self.folder = path_to_data
        self.outcome_name = outcome_name
        if outcome_name == 'cardio_disease':
            self._filename_prefix = ''
        else:
            self._filename_prefix = outcome_name + '_study_'
        self.stratified = stratified
        self.include_main_effect = include_main_effect
        self.include_interaction = include_interaction
        self.include_gender_age_interaction = include_gender_age_interaction
        self.stratified_by_menopause = stratified_by_menopause
        self.exclude_subgroup_indicator = exclude_subgroup_indicator
        self.n_strata = n_strata
        self.baseline_strata = baseline_strata
        self.treatment_encoding = treatment_encoding

    @property
    def n_age_groups(self):
        _, _, age_names \
            = self.get_age_and_gender_indicator_among_binary_covariates()
        return len(age_names)

    @property
    def n_age_gender_interaction(self):
        interaction_included = int(
            self.include_gender_age_interaction or self.stratified_by_menopause
        )
        return interaction_included * 2 * self.n_age_groups

    @property
    def n_fixed_effect(self):
        return (
            1 + self.stratified * (self.n_strata - 1)
            + self.n_age_gender_interaction
        )

    def read_legend_outcome(self, type='survival_time'):
        """
        Parameters
        ----------
        type : str, {'treatment', 'survival_time'}
        """

        if type == 'treatment':
            outcome = self.read_treatment_indicator(self.treatment_encoding)

        elif type == 'survival_time':
            event_time = np.genfromtxt(
                self.folder + self._filename_prefix + 'event_time.txt',
                missing_values='NA', filling_values=np.inf
            )
            censoring_time = np.genfromtxt(
                self.folder + self._filename_prefix + 'censoring_time.txt',
                missing_values='NA', filling_values=np.inf
            )
            outcome = (event_time, censoring_time)

        else:
            raise ValueError()

        return outcome

    def read_legend_design_matrix(self, matrix_format='csr'):
        """ Read the (pre-processed) design matrix. """

        X = self.read_legend_predictors()
        treatment = self.read_treatment_indicator(self.treatment_encoding)
        treatment = sp.sparse.csc_matrix(treatment[:, np.newaxis])

        if self.include_interaction:
            X_interaction = self.create_interaction(treatment, X)

        if not self.include_main_effect:
            X = sp.sparse.csc_matrix((X.shape[0], 0))

        if self.include_gender_age_interaction or self.stratified_by_menopause:
            X_age_gender_main = self.read_age_and_gender_interaction()
            X_age_gender_interaction = self.create_interaction(
                treatment, X_age_gender_main
            )
            X = sp.sparse.hstack((
                X_age_gender_main, X_age_gender_interaction, X
            ))

        if self.include_interaction:
            X = sp.sparse.hstack((X, X_interaction))

        if self.stratified:
            strata_id = self.read_propensity_strata()
            X = sp.sparse.hstack((strata_id, X))

        X = sp.sparse.hstack((treatment, X))
        return X.asformat(matrix_format)

    def read_age_and_gender_interaction(self):
        X_bin = self.read_sp_coo(
            self.folder + self._filename_prefix + 'sparse_design_matrix_binary_part.txt')
        X_bin = X_bin.tocsc()
        age_indicator, gender_indicator, age_name \
            = self.get_age_and_gender_indicator_among_binary_covariates()
        X_gender = X_bin[:, gender_indicator]
        if self.stratified_by_menopause:
            X_menopause_status = sp.sparse.hstack([
                sp.sparse.csc_matrix(X_bin[:, indicator].sum(-1) > 0)
                for indicator in age_indicator
            ])
            X_age_gender = self.create_interaction(
                X_gender, X_menopause_status
            )
        else:
            X_age_gender = self.create_interaction(
                X_gender, X_bin[:, age_indicator]
            )
        return X_age_gender

    def get_age_and_gender_indicator_among_binary_covariates(
            self, sorted_by_age=True):
        binary_covariate_names = np.loadtxt(
            self.folder + self._filename_prefix + 'covariate_name_binary_part.txt',
            dtype=np.str_, delimiter='\n'
        )
        gender_indicator = self.search_covariate_indices(
            'gender = FEMALE', exclude_words='Subgroup',
            covariate_name=binary_covariate_names
        )
        age_indicator = self.search_covariate_indices(
            'age group', covariate_name=binary_covariate_names
        )
        age_name = binary_covariate_names[age_indicator]
        if sorted_by_age:
            sort_index = np.argsort([name.split(': ')[-1] for name in age_name])
            age_indicator = age_indicator[sort_index]
            age_name = age_name[sort_index]
        if self.stratified_by_menopause:
            age_ranges = self.get_age_range(age_name)
            menopause_onset = 45
            menopause_end = 55
            pre_menopause_indicator = age_indicator[[
                age_range[-1] < menopause_onset for age_range in age_ranges
            ]]
            peri_menopause_indicator = age_indicator[[
                menopause_onset <= age_range[0] and age_range[1] < menopause_end
                for age_range in age_ranges
            ]]
            post_menopause_indicator = age_indicator[[
                menopause_end <= age_range[0]
                for age_range in age_ranges
            ]]
            age_indicator = [
                pre_menopause_indicator, peri_menopause_indicator, post_menopause_indicator
            ]
            age_name = ['pre_menopause', 'peri_menopause', 'post_menopause']
        return age_indicator, gender_indicator, age_name

    @staticmethod
    def get_age_range(age_name):
        return [np.array(name.split(': ')[-1].split('-'), dtype=np.int)
                for name in age_name]

    @staticmethod
    def create_interaction(covariate, X):
        """ Create interaction between one covariate and columns of a design matrix. """
        X = X.tocsc()
        X_interaction = sp.sparse.hstack(tuple(
            X[:, j].multiply(covariate)
            for j in range(X.shape[1])
        ))
        return X_interaction

    def read_treatment_indicator(self, encoding=None):
        """
        Parameters
        ----------
        encoding : {'ace', 'thiazide', None}
        """
        if encoding is None:
            encoding = self.treatment_encoding

        treatment = np.loadtxt(
            self.folder + self._filename_prefix + 'treatment.txt')
        if encoding == 'thiazide':
            treatment = 1 - treatment
        return treatment

    def read_legend_predictors(self, format='csr'):
        """ Return the design matrix for the clinical covariates (main effects). """
        X_bin = self.read_sp_coo(
            self.folder + self._filename_prefix + 'sparse_design_matrix_binary_part.txt')
        if self.exclude_subgroup_indicator:
            subgroup_col_indicator = self.get_subgroup_column_indicator()
            X_bin = X_bin.tocsc()[:, np.logical_not(subgroup_col_indicator)]
        X_cont = self.read_sp_coo(
            self.folder + self._filename_prefix + 'sparse_design_matrix_cont_part.txt')
        X = sp.sparse.hstack((X_bin, X_cont), format=format)

        return X

    def get_diabetes_subgroup_column_id(self):
        column_id_bin = np.loadtxt(
            self.folder + self._filename_prefix + 'covariate_id_binary_part.txt')
        return np.where(column_id_bin == 7998)[0]

    def get_subgroup_column_indicator(self, return_index=False):
        subgroup_id = [1998, 2998, 3998, 4998, 5998, 6998, 7998, 8998]
        column_id_bin = np.loadtxt(
            self.folder + self._filename_prefix + 'covariate_id_binary_part.txt')
        subgroup_col_indicator = \
            np.array([id in subgroup_id for id in column_id_bin])
        if return_index:
            subgroup_col_indicator = np.where(subgroup_col_indicator)[0]
        return subgroup_col_indicator

    def read_sp_coo(self, filename):
        A = pd.read_table(filename, delim_whitespace=True,
                          dtype={'i': np.int32, 'j': np.int32, 'val': np.float64})
        A = sp.sparse.coo_matrix((A.val, (A.i, A.j)))
        return A

    def read_propensity_strata(self):
        propensity_score = np.loadtxt(
            self.folder + self._filename_prefix + 'propensity_score.txt')
        strata_ind = DataManager.create_strata_indicator(
            propensity_score, self.n_strata, self.baseline_strata
        )
        return strata_ind

    def read_covariate_name(self):

        covariate_name = self.read_legend_predictor_names()

        if self.treatment_encoding == 'ace':
            drug_name = "ACE inhibitor"
        else:
            drug_name = "thiazide diuretics"

        if self.include_interaction:
            interaction_name = np.array([
                drug_name + " interaction with " + name for name in covariate_name
            ])

        if not self.include_main_effect:
            covariate_name = np.array([], dtype=np.unicode_)

        if self.include_gender_age_interaction or self.stratified_by_menopause:
            _, gender_indicator, age_names \
                = self.get_age_and_gender_indicator_among_binary_covariates()
            age_gender_main_name = np.array([
                'gender = FEMALE in ' + age_name
                for age_name in age_names
            ])
            age_gender_interaction_name = np.array([
                drug_name + " interaction with " + age_gender_name
                for age_gender_name in age_gender_main_name
            ])
            covariate_name = np.concatenate((
                age_gender_main_name, age_gender_interaction_name, covariate_name
            ))

        if self.include_interaction:
            covariate_name = np.concatenate((covariate_name, interaction_name))

        if self.stratified:
            strata_id_name = np.array([
                'propensity_strata_' + str(i + 1)
                for i in range(self.n_strata)
            ])
            strata_id_name = np.delete(strata_id_name, self.baseline_strata)
            covariate_name = np.concatenate((strata_id_name, covariate_name))

        covariate_name = np.concatenate((
            ['Treatment by ' + drug_name], covariate_name
        ))
        return covariate_name

    def read_legend_predictor_names(self):
        """ Return the names of the main effects. """
        cont_covariate_name = np.loadtxt(
            self.folder + self._filename_prefix + 'covariate_name_cont_part.txt',
            dtype=np.str_, delimiter='\n'
        )
        bin_convariate_name = np.loadtxt(
            self.folder + self._filename_prefix + 'covariate_name_binary_part.txt',
            dtype=np.str_, delimiter='\n'
        )
        if self.exclude_subgroup_indicator:
            subgroup_col_indicator = self.get_subgroup_column_indicator()
            bin_convariate_name = bin_convariate_name[
                np.logical_not(subgroup_col_indicator)
            ]
        covariate_name = np.concatenate((
            bin_convariate_name, cont_covariate_name
        ))
        return covariate_name

    def shorten_covariate_name(self, names):
        names = np.atleast_1d(names)
        names = np.array([
            self._shorten_singleton_covariate_name(name) for name in names
        ])
        if names.size == 1:
            names = names[0]
        return names

    @staticmethod
    def _shorten_singleton_covariate_name(name):
        emdash = '\u2014'
        return name.replace(
            'during day -365 through 0 days relative to index:', emdash
        ).replace(
            'during day -30 through 0 days relative to index:', emdash
        )

    def search_covariate_indices(
            self, keywords, exclude_words=None, covariate_name=None):
        if covariate_name is None:
            covariate_name = self.read_covariate_name()

        keywords = np.atleast_1d(keywords)
        keywords = np.array([word.lower() for word in keywords])
        if exclude_words is not None:
            exclude_words = np.atleast_1d(exclude_words)
            exclude_words = np.array([word.lower() for word in exclude_words])

        covariate_index = np.where(np.array([
            self.has_keyword(name.lower(), keywords, exclude_words)
            for name in covariate_name
        ]))[0]

        if covariate_index.size == 1:
            covariate_index = covariate_index.item()

        return covariate_index

    def get_gender_index(self):
        gender_index = self.search_covariate_indices(
            'gender = FEMALE', exclude_words='interaction'
        )
        treatment_gender_index = self.search_covariate_indices(
            [self.treatment_encoding, 'gender = FEMALE']
        )
        gender_age_index = self.search_covariate_indices(
            'gender = FEMALE interaction with age group'
        )
        return gender_index, treatment_gender_index, gender_age_index

    def has_keyword(self, string, keywords, exclude_words=None):
        """ Check that all the keywords matche and that none of
        the exclusion words match. """
        keyword_found = np.all([word in string for word in keywords])
        if exclude_words is None:
            exclude_word_found = False
        else:
            exclude_word_found = np.any([
                word in string for word in exclude_words
            ])
        return keyword_found and (not exclude_word_found)

    @staticmethod
    def create_strata_indicator(score, n_strata, baseline_strata):
        """
        Parameters
        ----------
        baseline_strata : int
            zero-based index
        """
        percent = np.linspace(0, 100, n_strata + 1)
        percentile = np.percentile(score, percent)
        percentile[0] -= 1  # Make sure the inequality behaves correctly.
        X = np.logical_and(
            percentile[np.newaxis, :-1] < score[:, np.newaxis],
            score[:, np.newaxis] <= percentile[np.newaxis, 1:]
        ).astype(np.float)
        X = np.delete(X, baseline_strata, axis=1)
        return sp.sparse.csr_matrix(X)



import numpy as np
import scipy as sp
import scipy.sparse
import scipy.linalg
from warnings import warn

class ConjugateGradientSampler():

    def __init__(self, n_coef_wo_shrinkage):
        self.n_coef_wo_shrinkage = n_coef_wo_shrinkage

    def sample(
            self, X, omega, prior_prec_sqrt, z,
            beta_init=None, precond_by='prior', beta_scaled_sd=None,
            maxiter=None, atol=10e-6, seed=None):
        """
        Generate a multi-variate Gaussian with the mean mu and covariance Sigma of the form
           Sigma^{-1} = X' Omega X + prior_prec_sqrt^2, mu = Sigma z
        where D is assumed to be diagonal. For numerical stability, the code first sample
        from the scaled parameter regress_coef / precond_scale.

        Param:
        ------
        D : vector
        atol : float
            The absolute tolerance on the residual norm at the termination
            of CG iterations.
        beta_scaled_sd : vector of length X.shape[1]
            Used to estimate a good preconditioning scale for the coefficient
            without shrinkage. Used only if precond_by == 'prior'.
        precond_by : {'prior', 'diag'}
        """

        if seed is not None:
            np.random.seed(seed)

        # Define a preconditioned linear operator.
        Phi_precond_op, precond_scale = \
            self.precondition_linear_system(
                prior_prec_sqrt, omega, X, precond_by, beta_scaled_sd
            )

        # Draw a target vector.
        v = X.Tdot(omega ** (1 / 2) * np.random.randn(X.shape[0])) \
            + prior_prec_sqrt * np.random.randn(X.shape[1])
        b = precond_scale * (z + v)

        # Callback function to count the number of PCG iterations.
        cg_info = {'n_iter': 0}
        def cg_callback(x): cg_info['n_iter'] += 1

        # Run PCG.
        rtol = atol / np.linalg.norm(b)
        beta_scaled_init = beta_init / precond_scale
        beta_scaled, info = sp.sparse.linalg.cg(
            Phi_precond_op, b, x0=beta_scaled_init, maxiter=maxiter, tol=rtol,
            callback=cg_callback
        )

        if info != 0:
            warn(
                "The conjugate gradient algorithm did not achieve the requested " +
                "tolerance level. You may increase the maxiter or use the dense " +
                "linear algebra instead."
            )

        beta = precond_scale * beta_scaled
        cg_info['valid_input'] = (info >= 0)
        cg_info['converged'] = (info == 0)

        return beta, cg_info

    def precondition_linear_system(
            self, prior_prec_sqrt, omega, X, precond_by, beta_scaled_sd):

        # Compute the preconditioners.
        precond_scale = self.choose_preconditioner(
            prior_prec_sqrt, omega, X, precond_by, beta_scaled_sd
        )

        # Define a preconditioned linear operator.
        precond_prior_prec = (precond_scale * prior_prec_sqrt) ** 2
        def Phi_precond(x):
            Phi_x = precond_prior_prec * x \
                    + precond_scale * X.Tdot(omega * X.dot(precond_scale * x))
            return Phi_x
        Phi_precond_op = sp.sparse.linalg.LinearOperator(
            (X.shape[1], X.shape[1]), matvec=Phi_precond
        )
        return Phi_precond_op, precond_scale

    def choose_preconditioner(
            self, prior_prec_sqrt, omega, X, precond_by, beta_scaled_sd):

        precond_scale = self.choose_diag_preconditioner(
            prior_prec_sqrt, omega, X, precond_by, beta_scaled_sd)

        return precond_scale

    def choose_diag_preconditioner(
            self, prior_prec_sqrt, omega, X, precond_by='diag',
            beta_scaled_sd=None):
        # Compute the diagonal (sqrt) preconditioner.

        if precond_by == 'prior':

            precond_scale = np.ones(len(prior_prec_sqrt))
            precond_scale[self.n_coef_wo_shrinkage:] = \
                prior_prec_sqrt[self.n_coef_wo_shrinkage:] ** -1
            if self.n_coef_wo_shrinkage > 0:
                target_sd_scale = 2.
                    # Larger than 1 because it is better to err on the side
                    # of introducing large precisions.
                precond_scale[:self.n_coef_wo_shrinkage] = \
                    target_sd_scale * beta_scaled_sd[:self.n_coef_wo_shrinkage]

        elif precond_by == 'diag':
            diag = prior_prec_sqrt ** 2 + X.compute_fisher_info(weight=omega, diag_only=True)
            precond_scale = 1 / np.sqrt(diag)

        elif precond_by is None:
            precond_scale = np.ones(X.shape[1])

        else:
            raise NotImplementedError()

        return precond_scale


import math
import time
from warnings import warn
import numpy as np


class SamplerOptions():

    def __init__(self, coef_sampler_type,
                 global_scale_update='sample',
                 hmc_curvature_est_stabilized=False):
        """
        Parameters
        ----------
        coef_sampler_type : {'cholesky', 'cg', 'hmc'}
        global_scale_update : str, {'sample', 'optimize', None}
        hmc_curvature_est_stabilized : bool
        """
        if coef_sampler_type not in ('cholesky', 'cg', 'hmc'):
            raise ValueError("Unsupported regression coefficient sampler.")
        self.coef_sampler_type = coef_sampler_type
        self.gscale_update = global_scale_update
        self.curvature_est_stabilized = hmc_curvature_est_stabilized

    def get_info(self):
        return {
            'coef_sampler_type': self.coef_sampler_type,
            'global_scale_update': self.gscale_update,
            'hmc_curvature_est_stabilized': self.curvature_est_stabilized
        }

    @staticmethod
    def create(coef_sampler_type, options, model_name, design):
        """ Initialize class with, if unspecified, an appropriate default
        sampling method based on the type and size of model.
        """
        if options is None:
            options = {}

        if 'coef_sampler_type' in options:
            if coef_sampler_type is not None:
                warn("Duplicate specification of method for sampling "
                     "regression coefficient. Will use the dictionary one.")
            coef_sampler_type = options['coef_sampler_type']

        if coef_sampler_type not in (None, 'cholesky', 'cg', 'hmc'):
            raise ValueError("Unsupported sampler type.")

        if model_name in ('linear', 'logit'):

            n_obs, n_pred = design.shape
            if not design.is_sparse:
                preferred_method = 'cholesky'
            else:
                # TODO: Make more informed choice between Cholesky and CG.
                frac = design.nnz / (n_obs * n_pred)
                fisher_info_cost = frac ** 2 * n_obs * n_pred ** 2
                cg_cost = design.nnz * 100.
                preferred_method = 'cg' if cg_cost < fisher_info_cost \
                    else 'cholesky'

            # TODO: Implement Woodbury-based Gaussian sampler.
            if n_pred > n_obs:
                warn("Sampler has not been optimized for 'small n' problem.")

            if coef_sampler_type is None:
                coef_sampler_type = preferred_method
            elif coef_sampler_type not in ('hmc', preferred_method):
                warn("Specified sampler may not be optimal. Worth experimenting "
                     "with the '{:s}' option.".format(preferred_method))

        else:
            if coef_sampler_type != 'hmc':
                warn("Specified sampler type is not supported for the {:s} "
                     "model. Will use HMC instead.".format(model_name))
            coef_sampler_type = 'hmc'

        options['coef_sampler_type'] = coef_sampler_type
        return SamplerOptions(**options)


class MarkovChainManager():

    def __init__(self, n_obs, n_pred, n_unshrunk, model_name):
        self.n_obs = n_obs
        self.n_pred = n_pred
        self.n_unshrunk = n_unshrunk
        self.model_name = model_name
        self._prev_timestamp = None # For status update during Gibbs
        self._curr_timestamp = None

    def merge_outputs(self, mcmc_output, next_mcmc_output):

        for output_key in ['samples', '_reg_coef_sampling_info']:
            curr_output = mcmc_output[output_key]
            next_output = next_mcmc_output[output_key]
            next_mcmc_output[output_key] = {
                key : np.concatenate(
                    (curr_output[key], next_output[key]), axis=-1
                ) for key in curr_output.keys()
            }

        next_mcmc_output['n_iter'] += mcmc_output['n_iter']
        next_mcmc_output['runtime'] += mcmc_output['runtime']

        for output_key in ['_init_optim_info', 'seed']:
            next_mcmc_output[output_key] = mcmc_output[output_key]

        return next_mcmc_output

    def pre_allocate(self, samples, sampling_info, n_post_burnin, thin, params_to_save, sampling_method):

        n_sample = math.floor(n_post_burnin / thin)  # Number of samples to keep

        if 'coef' in params_to_save:
            samples['coef'] = np.zeros((self.n_pred, n_sample))

        if 'local_scale' in params_to_save:
            samples['local_scale'] = np.zeros((self.n_pred - self.n_unshrunk, n_sample))

        if 'global_scale' in params_to_save:
            samples['global_scale'] = np.zeros(n_sample)

        if 'obs_prec' in params_to_save:
            if self.model_name == 'linear':
                samples['obs_prec'] = np.zeros(n_sample)
            elif self.model_name == 'logit':
                samples['obs_prec'] = np.zeros((self.n_obs, n_sample))

        if 'logp' in params_to_save:
            samples['logp'] = np.zeros(n_sample)

        for key in self.get_sampling_info_keys(sampling_method):
            sampling_info[key] = np.zeros(n_sample)

    def get_sampling_info_keys(self, sampling_method):
        if sampling_method == 'cg':
            keys = ['n_cg_iter']
        elif sampling_method in ['hmc', 'nuts']:
            keys = [
                'stepsize', 'n_hessian_matvec', 'n_grad_evals',
                'stability_limit_est', 'stability_adjustment_factor',
                'instability_detected'
            ]
            if sampling_method == 'hmc':
                keys += ['n_integrator_step', 'accepted', 'accept_prob']
            else:
                keys += ['tree_height', 'ave_accept_prob']
        else:
            keys = []
        return keys

    def store_current_state(
            self, samples, mcmc_iter, n_burnin, thin, coef, lscale,
            gscale, obs_prec, logp, params_to_save):

        if mcmc_iter <= n_burnin or (mcmc_iter - n_burnin) % thin != 0:
            return

        index = math.floor((mcmc_iter - n_burnin) / thin) - 1

        if 'coef' in params_to_save:
            samples['coef'][:, index] = coef

        if 'local_scale' in params_to_save:
            samples['local_scale'][:, index] = lscale

        if 'global_scale' in params_to_save:
            samples['global_scale'][index] = gscale

        if 'obs_prec' in params_to_save:
            if self.model_name == 'linear':
                samples['obs_prec'][index] = obs_prec
            elif self.model_name == 'logit':
                samples['obs_prec'][:, index] = obs_prec

        if 'logp' in params_to_save:
            samples['logp'][index] = logp

    def store_sampling_info(
            self, sampling_info, info, mcmc_iter, n_burnin, thin, sampling_method):

        if mcmc_iter <= n_burnin or (mcmc_iter - n_burnin) % thin != 0:
            return

        index = math.floor((mcmc_iter - n_burnin) / thin) - 1
        for key in self.get_sampling_info_keys(sampling_method):
            sampling_info[key][index] = info[key]

    def pack_parameters(self, coef, obs_prec, lscale, gscale):
        state = {
            'coef': coef,
            'local_scale': lscale,
            'global_scale': gscale,
        }
        if self.model_name in ('linear', 'logit'):
            state['obs_prec'] = obs_prec
        return state

    def stamp_time(self, curr_time):
        self._prev_timestamp = curr_time

    def print_status(self, n_status_update, mcmc_iter, n_iter,
                     time_format='minute'):

        if n_status_update == 0:
            return
        n_iter_per_update = int(n_iter / n_status_update)
        if mcmc_iter % n_iter_per_update != 0:
            return

        self._curr_timestamp = time.time()

        time_elapsed = self._curr_timestamp - self._prev_timestamp
        if time_format == 'second':
            time_str = "{:.3g} seconds".format(time_elapsed)
        elif time_format == 'minute':
            time_str = "{:.3g} minutes".format(time_elapsed / 60)
        else:
            raise ValueError()

        msg = " ".join((
            "{:d} Gibbs iterations complete:".format(mcmc_iter),
            time_str, "has elasped since the last update."
        ))
        print(msg)
        self._prev_timestamp = self._curr_timestamp


from math import exp, log, log10, sqrt, copysign
from .util import warn_message_only
import scipy.stats as stats


class HamiltonianBasedStepsizeAdapter():
    """
    Updates the stepsize of an HMC integrator so that the average Hamiltonian
    error matches a pre-specified target value.
    """

    def __init__(self, init_stepsize, target_accept_prob=.9,
                 init_adaptsize=1., adapt_decay_exponent=1.,
                 reference_iteration=500, adaptsize_at_reference=.05):
        """
        Parameters
        ----------
        reference_iteration & adaptsize_at_reference:
            Stepsize sequence of Robbins-Monro algorithm will be set so that it
            decreases to `adaptsize_at_refrence` after `reference_iteration`.
        """
        if init_stepsize <= 0:
            raise ValueError("The initial stepsize must be positive.")
        log_init_stepsize = log(init_stepsize)
        self.log_stepsize = log_init_stepsize
        self.log_stepsize_averaged = log_init_stepsize
        self.n_averaged = 0
        self.target_accept_prob = target_accept_prob
        self.target_log10_hamiltonian_error \
            = self.convert_to_log_hamiltonian_error(target_accept_prob)

        self.rm_stepsizer = RobbinsMonroStepsizer(
            init=init_adaptsize,
            decay_exponent=adapt_decay_exponent,
            reference_iteration=reference_iteration,
            size_at_reference=adaptsize_at_reference
        )

    @staticmethod
    def convert_to_log_hamiltonian_error(target_accept_prob):
        """ Calculate the target squared Hamiltonian error in the log scale.

        Under a high-dimensional limit of i.i.d. parameters, the Hamiltonian
        error is distributed as
            Normal(mean = - delta / 2, var = delta),
        and the corresponding average acceptance rate is
            2 GausssianCDF(- sqrt(delta) / 2).
        So we solve for `delta` that theoretically achieves the target acceptance
        rate and try to calibrate the average square error of the Hamiltonian
        to be the theoretical value (delta^2 / 4 + delta).
        """
        if target_accept_prob <= 0 or target_accept_prob >= 1:
            raise ValueError("Target probability must be within (0, 1).")
        delta = 4 * stats.norm.ppf(target_accept_prob / 2) ** 2
        target_log10_hamiltonian_error = .5 * log10(delta + delta ** 2 / 4)
        return target_log10_hamiltonian_error

    def get_current_stepsize(self, averaged=False):
        if averaged:
            return exp(self.log_stepsize_averaged)
        else:
            return exp(self.log_stepsize)

    def reinitialize(self, init_stepsize):
        log_init_stepsize = log(init_stepsize)
        self.log_stepsize = log_init_stepsize
        self.log_stepsize_averaged = log_init_stepsize
        self.n_averaged = 0

    def adapt_stepsize(self, hamiltonian_error):
        rm_stepsize = self.rm_stepsizer.calculate_stepsize(self.n_averaged)
        self.n_averaged += 1
        adaptsize = self.transform_to_adaptsize(hamiltonian_error)
        self.log_stepsize += rm_stepsize * adaptsize
        weight = 1 / self.n_averaged
        self.log_stepsize_averaged = (
            weight * self.log_stepsize
            + (1 - weight) * self.log_stepsize_averaged
        )
        return exp(self.log_stepsize)

    def transform_to_adaptsize(
            self, error, upper_bound=1., trans_type='piecewise'):
        """
        Parameters
        ----------
        trans_type: str, {'log-linear', 'sign', 'piecewise'}
        """

        if trans_type == 'probability':
            accept_prob = min(1, exp(error))
            adapt_size = accept_prob - self.target_accept_prob
            return adapt_size

        if error == 0.:
            log10_error = - float('inf')
        else:
            log10_error = log10(abs(error))

        target = self.target_log10_hamiltonian_error
        if trans_type == 'log-linear':
            adapt_size = target - log10_error

        elif trans_type == 'sign':
            adapt_size = copysign(1., target - log10_error)

        elif trans_type == 'piecewise':
            # Increase the adjustment when the error is larger than the target.
            if log10_error > target:
                adapt_size = (target - log10_error) / .301 # Convert to log2 scale.
            else:
                adapt_size = (target - log10_error) / 3 # Convert to log1000 scale.

        else:
            raise NotImplementedError()

        if abs(adapt_size) > upper_bound:
            adapt_size = copysign(1., adapt_size)

        return adapt_size


def initialize_stepsize(compute_acceptprob, dt=1.0):
    """ Heuristic for choosing an initial value of dt

    Parameters
    ----------
    compute_acceptprob: callable
        Computes the acceptance probability of the proposal one-step HMC proposal.
    """

    # Figure out what direction we should be moving dt.
    acceptprob = compute_acceptprob(dt)
    direc = 2 * int(acceptprob > 0.5) - 1

    # Keep moving dt in that direction until acceptprob crosses 0.5.
    while acceptprob == 0 or (2 * acceptprob) ** direc > 1:
        dt = dt * (2 ** direc)
        acceptprob = compute_acceptprob(dt)
        if acceptprob == 0 and direc == 1:
            # The last doubling of stepsize was too much.
            dt /= 2
            break

    return dt


class RobbinsMonroStepsizer():

    def __init__(self, init=1., decay_exponent=1.,
                 reference_iteration=None, size_at_reference=None):
        self.init = init
        self.exponent = decay_exponent
        self.scale = self.determine_decay_scale(
            init, decay_exponent, reference_iteration, size_at_reference
        )

    def determine_decay_scale(self, init, decay_exponent, ref_iter, size_at_ref):

        if (ref_iter is not None) and (size_at_ref is not None):
            decay_scale = \
                ref_iter / ((init / size_at_ref) ** (1 / decay_exponent) - 1)
        else:
            warn_message_only(
                'The default stepsize sequence tends to decay too quicky; '
                'consider manually setting the decay scale.'
            )
            decay_scale = 1.

        return decay_scale

    def __iter__(self):
        self.n_iter = 0
        return self

    def __next__(self):
        stepsize = self.calculate_stepsize(self.n_iter)
        self.n_iter += 1
        return stepsize

    def calculate_stepsize(self, n_iter):
        stepsize = self.init / (1 + n_iter / self.scale) ** self.exponent
        return stepsize


class RobbinsMonroStepsizeAdapter():

    def __init__(self, init_stepsize, target_accept_prob=.9,
                 init_adaptsize=1., adapt_decay_exponent=1.,
                 reference_iteration=100, adaptsize_at_reference=.05):
        """
        Parameters
        ----------
        reference_iteration & adaptsize_at_reference:
            Stepsize sequence of Robbins-Monro algorithm will be set so that it
            decreases to `adaptsize_at_refrence` after `reference_iteration`.
        """
        if init_stepsize <= 0:
            raise ValueError("The initial stepsize must be positive.")
        log_init_stepsize = log(init_stepsize)
        self.log_stepsize = log_init_stepsize
        self.log_stepsize_averaged = log_init_stepsize
        self.n_averaged = 0
        self.target_accept_prob = target_accept_prob

        self.rm_stepsizer = iter(RobbinsMonroStepsizer(
            init=init_adaptsize,
            decay_exponent=adapt_decay_exponent,
            reference_iteration=reference_iteration,
            size_at_reference=adaptsize_at_reference
        ))

    def get_current_stepsize(self, averaged=False):
        if averaged:
            return exp(self.log_stepsize_averaged)
        else:
            return exp(self.log_stepsize)

    def adapt_stepsize(self, accept_prob):
        self.n_averaged += 1
        rm_stepsize = next(self.rm_stepsizer)
        adaptsize = \
            self.transform_to_adaptsize(accept_prob, self.target_accept_prob)
        self.log_stepsize += rm_stepsize * adaptsize
        weight = 1 / self.n_averaged
        self.log_stepsize_averaged = (
            weight * self.log_stepsize
            + (1 - weight) * self.log_stepsize_averaged
        )
        return exp(self.log_stepsize)

    def transform_to_adaptsize(
            self, accept_prob, target, trans_type='linear'):
        """
        Parameters
        ----------
        trans_type: str, {'linear', 'sign', 'penalize-high-prob'}
        """

        if trans_type == 'linear':
            adapt_size = accept_prob - target

        elif trans_type == 'sign':
            adapt_size = copysign(1., accept_prob - target)

        elif trans_type == 'penalize-high-prob':
            # Transforms accept_prob -> adapt_size so that it roughly interpolate
            # the points (0, -1), (target, 0), and (1, 1). Transformation is
            # linear near accept_prob = target but quickly goes up to
            # adapt_size = 1 as (1 - accecpt_prob) becomes an order of manitude
            # smaller than (1 - target).
            if accept_prob <= target:
                adapt_size = (accept_prob - target) / target
            else:
                epsilon = 2. ** -52
                magnitude_diff = log10(
                    (1. - (accept_prob - epsilon)) / (1 - target)
                )
                if magnitude_diff == 0:
                    w = 0.
                else:
                    w = exp(magnitude_diff ** - 1)
                adapt_size = (
                    (1 - w) * (accept_prob - target) / target
                    - w * magnitude_diff
                )
                adapt_size = min(1., adapt_size)

        else:
            raise NotImplementedError()

        return adapt_size


class DualAverageStepsizeAdapter():

    def __init__(self, init_stepsize, target_accept_prob=.9):

        if init_stepsize <= 0:
            raise ValueError("The initial stepsize must be positive.")
        log_init_stepsize = log(init_stepsize)
        self.log_stepsize = log_init_stepsize
        self.log_stepsize_averaged = log_init_stepsize
        self.n_averaged = 0
        self.target_accept_prob = target_accept_prob
        self.latent_stat = 0.  # Used for dual-averaging.

        # Parameters for the dual-averaging algorithm.
        self.stepsize_averaging_log_decay_rate = 0.75
        self.latent_prior_samplesize = 10
        multiplier = 2. # > 1 to err on the side of shrinking toward a larger value.
        self.log_stepsize_shrinkage_mean = log(multiplier) + log_init_stepsize
        self.log_stepsize_shrinkage_strength = 0.05
            # Variable name is not quite accurate since this parameter interacts with latent_prior_samplesize.

    def get_current_stepsize(self, averaged=False):
        if averaged:
            return exp(self.log_stepsize_averaged)
        else:
            return exp(self.log_stepsize)

    def adapt_stepsize(self, accept_prob):
        self.n_averaged += 1
        self.latent_stat = self.update_latent_stat(
            accept_prob, self.target_accept_prob, self.latent_stat
        )
        self.log_stepsize, self.log_stepsize_averaged = self.dual_average_stepsize(
            self.latent_stat, self.log_stepsize_averaged
        )
        return exp(self.log_stepsize)

    def update_latent_stat(self, accept_prob, target_accept_prob, latent_stat):
        weight_latent = (self.n_averaged + self.latent_prior_samplesize) ** -1
        latent_stat = (1 - weight_latent) * latent_stat \
                      + weight_latent * (target_accept_prob - accept_prob)
        return latent_stat

    def dual_average_stepsize(self, latent_stat, log_stepsize_optimized):
        log_stepsize = (
            self.log_stepsize_shrinkage_mean
            - sqrt(self.n_averaged) / self.log_stepsize_shrinkage_strength * latent_stat
        )
        weight = self.n_averaged ** - self.stepsize_averaging_log_decay_rate
        log_stepsize_optimized = \
            (1 - weight) * log_stepsize_optimized + weight * log_stepsize
        return log_stepsize, log_stepsize_optimized


from setuptools import setup, find_packages

setup(
    name='mcmc_diagnostics',
    version='0.1',
    description='Python analogue of the R coda package',
    url='https://github.com/aki-nishimura/mcmc-diagnostics',
    author='Akihiko (Aki) Nishimura',
    author_email='akihiko4@ucla.edu',
    license='MIT',
    packages=find_packages()
)


from .abstract_model import AbstractModel
import math
import numpy as np


class LinearModel(AbstractModel):

    def __init__(self, y, design):
        self.y = y
        self.design = design
        self.name = 'linear'

    def compute_loglik_and_gradient(self, beta, obs_prec, loglik_only=False):
        X_beta = self.design.dot(beta)
        loglik = (
            len(self.y) * math.log(obs_prec) / 2
            - obs_prec * np.sum((self.y - X_beta) ** 2) / 2
        )
        if loglik_only:
            grad = None
        else:
            grad = obs_prec * self.design.Tdot(self.y - X_beta)
        return loglik, grad

    def compute_hessian(self, beta):
        pass

    def get_hessian_matvec_operator(self, beta, obs_prec):
        hessian_op = lambda v: - obs_prec * self.design.Tdot(self.design.dot(v))
        return hessian_op

    def calc_intercept_mle(self):
        return self.y.mean()

    @staticmethod
    def simulate_outcome(X, beta, noise_sd, seed=None):
        """
        Parameters
        ----------
        X : DesignMatrix, numpy/scipy matrix
            Only needs to support the `dot()` operation
        """
        if seed is not None:
            np.random.seed(seed)
        y = X.dot(beta) + noise_sd * np.random.randn(X.shape[0])
        return y


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.pylab import MaxNLocator


def plot_conf_interval(
        coef_samples, conf_level=.95, n_coef_to_plot=None,
        sort_by_median_val=False, marker_scale=1.0,
    ):
    tail_prob = (1 - conf_level) / 2
    lower, median, upper = [
        np.quantile(coef_samples, q, axis=-1)
        for q in [tail_prob, .5, 1 - tail_prob]
    ]

    if sort_by_median_val:
        sort_ind = np.argsort(median)
    else:
        sort_ind = np.arange(len(median))  # No sorting

    if n_coef_to_plot is None:
        n_coef_to_plot = len(median)
    coef_index = sort_ind[:n_coef_to_plot]

    plt.plot(
        coef_index, median[coef_index],
        'x', color='tab:blue', ms=marker_scale * 10,
        label='Posterior median'
    )
    plt.plot(
        coef_index, lower[coef_index],
        '_', color='tab:green', ms=marker_scale * 12, lw=marker_scale * 1.2,
        label='{:.1f}% credible interval'.format(100 * conf_level)
    )
    plt.plot(
        coef_index, upper[coef_index],
        '_', color='tab:green', ms=marker_scale * 12, lw=marker_scale * 1.2
    )
    plt.gca().get_xaxis().set_major_locator(MaxNLocator(integer=True))

    plotted_quantity = {
        'lower': lower[coef_index],
        'median': median[coef_index],
        'upper': upper[coef_index],
        'coef_index': coef_index
    }
    return plotted_quantity


"""
Fits an auto-regressive process to a univariate time series. The order is
determined by minimizing AIC and the coefficients are estimated via the
Yule-Walker equation.

The code is based on the function
    statsmodels.regression.linear_model.yule_walker
from the statsmodels package and has been adapted by Aki Nishimura.
"""

# License for the original Statsmodels code (BSD 3-clause)
# --------------------------------------------------------
#
# Copyright (C) 2006, Jonathan E. Taylor
# All rights reserved.
#
# Copyright (c) 2006-2008 Scipy Developers.
# All rights reserved.
#
# Copyright (c) 2009-2018 Statsmodels Developers.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of Statsmodels nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL STATSMODELS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.


import numpy as np
import math
from scipy.linalg import solve_toeplitz


def fit(x, max_order=None):

    if max_order is None:
        max_order = min(
            len(x) - 1, math.ceil(10 * np.log10(len(x)))
        )

    aic_best = float('inf')
    ar_order_best = 0
    ar_coef_best = None
    for order in range(max_order + 1):
        ar_coef, aic = fit_via_yule_walker(x, order)
        if aic < aic_best:
            aic_best = aic
            ar_order_best = order
            ar_coef_best = ar_coef

    return ar_order_best, ar_coef_best


def fit_via_yule_walker(x, order, acf_method="mle", demean=True):
    """
    Estimate AR(p) parameters of a sequence x using the Yule-Walker equation.

    Parameters
    ----------
    x : 1d numpy array
    order : integer
        The order of the autoregressive process.
    acf_method : {'unbiased', 'mle'}, optional
       Method can be 'unbiased' or 'mle' and this determines denominator in
       estimating autocorrelation function (ACF) at lag k. If 'mle', the
       denominator is  `n = x.shape[0]`, if 'unbiased' the denominator is `n - k`.
    demean : bool
        True, the mean is subtracted from `x` before estimation.
    """

    if demean:
        x = x.copy()
        x -= x.mean()

    if acf_method == "unbiased":
        denom = lambda lag: len(x) - lag
    else:
        denom = lambda lag: len(x)
    if x.ndim > 1 and x.shape[1] != 1:
        raise ValueError("expecting a vector to estimate AR parameters")

    auto_cov = np.zeros(order + 1, np.float64)
    auto_cov[0] = (x ** 2).sum() / denom(0)
    for lag in range(1, order + 1):
        auto_cov[lag] = np.sum(x[0:-lag] * x[lag:]) / denom(lag)

    if order == 0:
        ar_coef = None
        innovation_var = auto_cov[0]
    else:
        ar_coef = _solve_yule_walker(auto_cov)
        innovation_var = auto_cov[0] - (auto_cov[1:] * ar_coef).sum()

    aic = compute_aic(innovation_var, order, len(x))

    return ar_coef, aic


def _solve_yule_walker(auto_cov):
    ar_coef = solve_toeplitz(auto_cov[:-1], auto_cov[1:])
    return ar_coef


def compute_aic(innovation_var, ar_order, n_obs):
    # I don't quite understand the formula (the maximized likelihood part), but
    # it agrees with the one used in R's ar.yw.default as well as Python's statsmodels.
    # Also, described in http://pages.stern.nyu.edu/~churvich/TimeSeries/Handouts/AICC.pdf
    return n_obs * np.log(innovation_var) + 2 * (1 + ar_order)




import numpy as np
import math
import time
from .util import simplify_warnings # Monkey patch the warning format
from warnings import warn
from .random import BasicRandom
from .reg_coef_sampler import SparseRegressionCoefficientSampler
from .model import LogisticModel
from .prior import RegressionCoefPrior
from .gibbs_util import MarkovChainManager, SamplerOptions


class BayesBridge():
    """ Implement Gibbs sampler for Bayesian bridge sparse regression. """

    def __init__(self, model, prior=RegressionCoefPrior()):
        """
        Parameters
        ----------
        model : RegressionModel object
        prior : RegressionCoefPrior object
        """

        self.n_obs = model.n_obs
        self.n_pred = model.n_pred
        self.n_unshrunk = prior.n_fixed
        self.prior_sd_for_unshrunk = prior.sd_for_fixed.copy()
        if model.intercept_added:
            self.n_unshrunk += 1
            self.prior_sd_for_unshrunk = np.concatenate((
                [prior.sd_for_intercept], self.prior_sd_for_unshrunk
            ))

        self.model = model
        self.prior = prior
        self.rg = BasicRandom()
        self.manager = MarkovChainManager(
            self.n_obs, self.n_pred, self.n_unshrunk, model.name
        )

    # TODO: write a test to ensure that the output when resuming the Gibbs
    # sampler coincide with that without interruption.
    def gibbs_additional_iter(
            self, mcmc_output, n_add_iter, n_status_update=0,
            merge=False, deallocate=False):
        """ Resume Gibbs sampler from the last state.

        Parameter
        ---------
        mcmc_output : dict
            Output of a previous call to the 'gibbs' method.
        n_iter : int
        n_status_update : int
        merge : bool
            If True, merge the Gibbs sampler outputs from the previous and
            current runs and then return.
        deallocate : bool
            If True, clear the samples from the previous Gibbs run to save
            memory.

        Returns
        -------
        next_mcmc_output : dict
        """

        if merge and deallocate:
            warn(
                "To merge the outputs, the previous one cannot be deallocated.")
            deallocate = False

        self.rg.set_state(mcmc_output['_random_gen_state'])

        init = mcmc_output['_markov_chain_state']
        thin, bridge_exp, coef_sampler_type = (
            mcmc_output[key]
            for key in ['thin', 'bridge_exponent', 'coef_sampler_type']
        )
        curvature_est_stabilized = mcmc_output['options']['hmc_curvature_est_stabilized']
        params_to_save = mcmc_output['samples'].keys()

        # Initalize the regression coefficient sampler with the previous state.
        self.reg_coef_sampler = SparseRegressionCoefficientSampler(
            self.n_pred, self.prior_sd_for_unshrunk, coef_sampler_type,
            curvature_est_stabilized, self.prior.slab_size
        )
        self.reg_coef_sampler.set_internal_state(mcmc_output['_reg_coef_sampler_state'])

        if deallocate:
            mcmc_output['samples'].clear()

        next_mcmc_output = self.gibbs(
            n_add_iter, 0, thin, init=init,
            params_to_save=params_to_save,
            n_status_update=n_status_update,
            options=mcmc_output['options'],
            _add_iter_mode=True
        )
        if merge:
            next_mcmc_output \
                = self.manager.merge_outputs(mcmc_output, next_mcmc_output)

        return next_mcmc_output

    def gibbs(self, n_iter, n_burnin=0, thin=1, seed=None,
              init={}, params_to_save=('coef', 'global_scale', 'logp'),
              coef_sampler_type=None, n_status_update=0,
              options=None, _add_iter_mode=False):
        """ Sample from the posterior under the specified model and prior.

        Parameters
        ----------
        n_iter : int
            total number of MCMC iterations i.e. burn-ins + saved posterior draws
        n_burnin : int
            number of burn-in samples to be discarded
        coef_sampler_type : {None, 'cholesky', 'cg', 'hmc'}
            Specifies the sampling method used to update regression coefficients.
            If None, the method is chosen via a crude heuristic based on the
            model type, as well as size and sparsity level of design matrix.
            For linear and logistic models with large and sparse design matrix,
            the conjugate gradient sampler ('cg') is preferred over the
            Cholesky decomposition based sampler ('cholesky'). For other
            models, only Hamiltonian Monte Carlo ('hmc') can be used.
        params_to_save : {'all', tuple or list of str}
            Specifies which parameters to save during MCMC iterations. If None,
            the most relevant parameters --- regression coefficients,
            global scale, posterior log-density --- are saved. Use all to save
            all the parameters (but beaware of the extra memory requirement),
            including local scale and, depending on the model, precision (
            inverse variance) of observations.
        n_status_update : int
            Number of updates to print on stdout during the sampler run.

        Other Parameters
        ----------------
        options : None, dict, SamplerOptions
            SamplerOptions class or a dict whose keywords are used as inputs
            to the class.

        Returns
        -------
        mcmc_output : dict
            Contains posterior samples under the key 'samples,' along with the
            sampler settings to reproduce and resume the sampling process.
        samples = mcmc_output['samples'] : dict of numpy arrays
            Contains MCMC samples of the parameters as specified by
            **params_to_save**. The last dimension of the arrays correspond
            to MCMC iterations; for example,
            :code:`samples['coef'][:, 0]`
            is the first MCMC sample of regression coefficients.
        """

        if not isinstance(options, SamplerOptions):
            options = SamplerOptions.create(
                coef_sampler_type, options, self.model.name, self.model.design
            )

        if not _add_iter_mode:
            self.rg.set_seed(seed)
            self.reg_coef_sampler = SparseRegressionCoefficientSampler(
                self.n_pred, self.prior_sd_for_unshrunk,
                options.coef_sampler_type, options.curvature_est_stabilized,
                self.prior.slab_size
            )

        if params_to_save == 'all':
            params_to_save = (
                'coef', 'local_scale', 'global_scale', 'logp'
            )
            if self.model.name != 'cox':
                params_to_save += ('obs_prec', )

        n_status_update = min(n_iter, n_status_update)
        start_time = time.time()
        self.manager.stamp_time(start_time)

        # Initial state of the Markov chain
        coef, obs_prec, lscale, gscale, init, initial_optim_info = \
            self.initialize_chain(init, self.prior.bridge_exp)

        # Pre-allocate
        samples = {}
        sampling_info = {}
        self.manager.pre_allocate(
            samples, sampling_info, n_iter - n_burnin, thin, params_to_save,
            options.coef_sampler_type
        )

        # Start Gibbs sampling
        for mcmc_iter in range(1, n_iter + 1):

            coef, info = self.update_regress_coef(
                coef, obs_prec, gscale, lscale, options.coef_sampler_type
            )

            obs_prec = self.update_obs_precision(coef)

            # Draw from gscale | coef and then lscale | gscale, coef.
            # (The order matters.)
            gscale = self.update_global_scale(
                gscale, coef[self.n_unshrunk:], self.prior.bridge_exp,
                method=options.gscale_update
            )

            lscale = self.update_local_scale(
                gscale, coef[self.n_unshrunk:], self.prior.bridge_exp)

            logp = self.compute_posterior_logprob(
                coef, gscale, obs_prec, self.prior.bridge_exp
            )

            self.manager.store_current_state(
                samples, mcmc_iter, n_burnin, thin, coef, lscale, gscale,
                obs_prec, logp, params_to_save
            )
            self.manager.store_sampling_info(
                sampling_info, info, mcmc_iter, n_burnin, thin,
                options.coef_sampler_type
            )
            self.manager.print_status(n_status_update, mcmc_iter, n_iter)

        runtime = time.time() - start_time

        if self.prior._gscale_paramet == 'coef_magnitude':
            gscale, lscale = \
                self.prior.adjust_scale(gscale, lscale, to='coef_magnitude')
            gscale_samples = samples.get('global_scale', 0.)
            lscale_samples = samples.get('local_scale', 0.)
            self.prior.adjust_scale(
                gscale_samples, lscale_samples, to='coef_magnitude'
            ) # Modify in place.

        _markov_chain_state = \
            self.manager.pack_parameters(coef, obs_prec, lscale, gscale)

        _reg_coef_sampling_info = None
        mcmc_output = {
            'samples': samples,
            'init': init,
            'n_iter': n_iter,
            'n_burnin': n_burnin,
            'thin': thin,
            'seed': seed,
            'n_coef_wo_shrinkage': self.n_unshrunk,
            'prior_sd_for_unshrunk': self.prior_sd_for_unshrunk,
            'bridge_exponent': self.prior.bridge_exp,
            'coef_sampler_type': options.coef_sampler_type,
            'runtime': runtime,
            'options': options.get_info(),
            '_init_optim_info': initial_optim_info,
            '_reg_coef_sampling_info': sampling_info,
            '_markov_chain_state': _markov_chain_state,
            '_random_gen_state': self.rg.get_state(),
            '_reg_coef_sampler_state': self.reg_coef_sampler.get_internal_state()
        }

        return mcmc_output

    def initialize_chain(self, init, bridge_exp):
        """ Choose the user-specified state if provided, the default ones otherwise."""

        valid_param_name \
            = ('coef', 'local_scale', 'global_scale', 'obs_prec', 'logp')
        for key in init:
            if key not in valid_param_name:
                warn("'{:s}' is not a valid parameter name and "
                     "will be ignored.".format(key))

        coef_only_specified = 'coef' in init and ('global_scale' not in init)

        if 'coef' in init:
            coef = init['coef']
            if not len(coef) == self.n_pred:
                raise ValueError('Invalid initial length of regression coefficient.')
        else:
            coef = np.zeros(self.n_pred)
            if self.model.name in ('linear', 'logit'):
                coef[0] = self.model.calc_intercept_mle()

        obs_prec = self.initialize_obs_precision(init, coef)

        if coef_only_specified:
            gscale = self.update_global_scale(
                None, coef[self.n_unshrunk:], bridge_exp,
                method='optimize'
            )
            lscale = self.update_local_scale(
                gscale, coef[self.n_unshrunk:], bridge_exp
            )
        else:
            if 'global_scale' in init:
                gscale = init['global_scale']
            else:
                gscale = self._get_default_global_scale(bridge_exp)
            if 'local_scale' in init:
                lscale = init['local_scale']
                if not len(lscale) == (self.n_pred - self.n_unshrunk):
                    raise ValueError('Invalid initial length of local scale parameter')
            else:
                lscale = np.ones(self.n_pred - self.n_unshrunk) / gscale

        if self.prior._gscale_paramet == 'coef_magnitude':
            # Gibbs sampler requires the raw parametrization, though
            # technically only gscale * lscale matters due to the update order.
            gscale, lscale \
                = self.prior.adjust_scale(gscale, lscale, to='raw')

        if 'coef' not in init:
            # Optimize coefficients and then update other parameters
            coef, info = self.reg_coef_sampler.search_mode(
                coef, lscale, gscale, obs_prec, self.model
            )
            obs_prec = self.update_obs_precision(coef)
            lscale = self.update_local_scale(
                gscale, coef[self.n_unshrunk:], bridge_exp
            )
            optim_info = {
                key: info[key] for key in ['is_success', 'n_design_matvec', 'n_iter']
            }
        else:
            optim_info = None

        init = {
            'coef': coef,
            'obs_prec': obs_prec,
            'local_scale': lscale,
            'global_scale': gscale
        }

        return coef, obs_prec, lscale, gscale, init, optim_info

    def _get_default_global_scale(self, bridge_exp):
        gscale_default = .1
        if self.prior._gscale_paramet == 'raw':
            gscale_default \
                /= self.prior.compute_power_exp_ave_magnitude(bridge_exp)
        return gscale_default

    def initialize_obs_precision(self, init, coef):
        if 'obs_prec' in init:
            obs_prec = np.ascontiguousarray(init['obs_prec'])
            # Cython requires a C-contiguous array.
            if not len(obs_prec) == self.n_obs:
                raise ValueError('An invalid initial state.')
        elif self.model.name == 'linear':
            obs_prec = np.mean(
                (self.model.y - self.model.design.dot(coef)) ** 2) ** -1
        elif self.model.name == 'logit':
            obs_prec = LogisticModel.compute_polya_gamma_mean(
                self.model.n_trial, self.model.design.dot(coef)
            )
        else:
            obs_prec = None
        return obs_prec

    def update_regress_coef(self, coef, obs_prec, gscale, lscale, sampling_method):

        if sampling_method in ('cholesky', 'cg'):

            if self.model.name == 'linear':
                y_gaussian = self.model.y
                obs_prec = obs_prec * np.ones(self.n_obs)
            elif self.model.name == 'logit':
                y_gaussian = (self.model.n_success - self.model.n_trial / 2) / obs_prec

            coef, info = self.reg_coef_sampler.sample_gaussian_posterior(
                y_gaussian, self.model.design, obs_prec, gscale, lscale,
                sampling_method
            )

        elif sampling_method in ['hmc', 'nuts']:
            coef, info = self.reg_coef_sampler.sample_by_hmc(
                coef, gscale, lscale, self.model, method=sampling_method
            )

        else:
            raise NotImplementedError()

        return coef, info

    def update_obs_precision(self, coef):

        obs_prec = None
        if self.model.name == 'linear':
            resid = self.model.y - self.model.design.dot(coef)
            scale = np.sum(resid ** 2) / 2
            obs_var = scale / self.rg.np_random.gamma(self.n_obs / 2, 1)
            obs_prec = 1 / obs_var
        elif self.model.name == 'logit':
            obs_prec = self.rg.polya_gamma(
                self.model.n_trial.astype(np.intc), self.model.design.dot(coef)
            )

        return obs_prec

    def update_global_scale(
            self, gscale, beta_with_shrinkage, bridge_exp,
            coef_expected_magnitude_lower_bd=.001, method='sample'):
        # :param method: {"sample", "optimize", None}

        if beta_with_shrinkage.size == 0:
            return 1. # arbitrary float value as a placeholder

        lower_bd = coef_expected_magnitude_lower_bd \
                   / self.prior.compute_power_exp_ave_magnitude(bridge_exp)
            # Solve for the value of global shrinkage such that
            # (expected value of regress_coef given gscale) = coef_expected_magnitude_lower_bd.

        if method == 'optimize':
            gscale = self.monte_carlo_em_global_scale(
                beta_with_shrinkage, bridge_exp)

        elif method == 'sample':
            # Conjugate update for phi = 1 / gscale ** bridge_exp
            if np.count_nonzero(beta_with_shrinkage) == 0:
                gscale = 0
            else:
                prior_param = self.prior.param['gscale_neg_power']
                shape, rate = prior_param['shape'], prior_param['rate']
                shape += beta_with_shrinkage.size / bridge_exp
                rate += np.sum(np.abs(beta_with_shrinkage) ** bridge_exp)
                phi = self.rg.np_random.gamma(shape, scale=1 / rate)
                gscale = 1 / phi ** (1 / bridge_exp)

        if (method is not None) and gscale < lower_bd:
            gscale = lower_bd
            warn(
                "The global shrinkage parameter update returned an unreasonably "
                "small value. Returning a specified lower bound value instead."
            )

        return gscale

    def monte_carlo_em_global_scale(
            self, beta_with_shrinkage, bridge_exp):
        """ Maximize the likelihood (not posterior conditional) 'coef | gscale'. """
        phi = len(beta_with_shrinkage) / bridge_exp \
              / np.sum(np.abs(beta_with_shrinkage) ** bridge_exp)
        gscale = phi ** - (1 / bridge_exp)
        return gscale

    def update_local_scale(self, gscale, beta_with_shrinkage, bridge_exp):

        lscale_sq = .5 / self.rg.tilted_stable(
            bridge_exp / 2, (beta_with_shrinkage / gscale) ** 2
        )
        lscale = np.sqrt(lscale_sq)

        # TODO: Pick the lower and upper bound more carefully.
        if np.any(lscale == 0):
            warn(
                "Local scale parameter under-flowed. Replacing with a small number.")
            lscale[lscale == 0] = 10e-16
        elif np.any(np.isinf(lscale)):
            warn(
                "Local scale parameter over-flowed. Replacing with a large number.")
            lscale[np.isinf(lscale)] = 2.0 / gscale

        return lscale

    def compute_posterior_logprob(self, coef, gscale, obs_prec, bridge_exp):

        # Contributions from the likelihood.
        params = [coef] if self.model.name != 'linear' else [coef, obs_prec]
        loglik, _ = self.model.compute_loglik_and_gradient(*params, loglik_only=True)

        # Contributions from the regularization.
        loglik += - .5 * np.sum((coef / self.prior.slab_size) ** 2)

        # Add contributions from the priors.
        prior_logp = 0
        n_shrunk_coef = len(coef) - self.n_unshrunk

        # for coef | gscale.
        prior_logp += \
            - n_shrunk_coef * math.log(gscale) \
            - np.sum(np.abs(coef[self.n_unshrunk:] / gscale) ** bridge_exp)

        # for coefficients without shrinkage.
        prior_logp += - 1 / 2 * np.sum(
            (coef[:self.n_unshrunk] / self.prior_sd_for_unshrunk) ** 2
        )
        prior_logp += - np.sum(np.log(
            self.prior_sd_for_unshrunk[self.prior_sd_for_unshrunk < float('inf')]
        ))
        prior_param = self.prior.param['gscale_neg_power']
        prior_logp += (prior_param['shape'] - 1.) * math.log(gscale) \
                      - prior_param['rate'] * gscale

        logp = loglik + prior_logp

        return logp



cimport cython
from libc.math cimport exp as exp_c
from libc.math cimport fabs, pow, log, sqrt, sin, floor, INFINITY, M_PI
import random
import numpy as np
cimport numpy as np
from numpy.random import PCG64
from numpy.random.bit_generator cimport BitGenerator
from bayesbridge.random.normal.normal cimport random_normal
from bayesbridge.random.uniform.uniform cimport random_uniform


cdef double MAX_EXP_ARG = 709  # ~ log(2 ** 1024)
ctypedef np.uint8_t np_uint8
ctypedef double (*rand_generator)()


cdef double exp(double x):
    if x > MAX_EXP_ARG:
        val = INFINITY
    elif x < - MAX_EXP_ARG:
        val = 0.
    else:
        val = exp_c(x)
    return val


@cython.cdivision(True)
cdef double sinc(double x):
    cdef double x_sq
    if fabs(x) < .01:
        x_sq = x * x
        val = 1. - x_sq / 6. * (1 - x_sq / 20.)
            # Taylor approximation with an error bounded by 2e-16
    else:
        val = sin(x) / x
    return val


cdef double python_builtin_next_double():
    return <double>random.random()


cdef class ExpTiltedStableDist():
    cdef rand_generator next_double
    cdef double TILT_POWER_THRESHOLD # For deciding the faster of two algorithms
    cdef BitGenerator bitgen

    def __init__(self, seed=None):
        self.set_seed(seed)
        self.bitgen = PCG64(seed)
        self.TILT_POWER_THRESHOLD = 2.

    def set_seed(self, seed):
        self.bitgen = PCG64(seed)

    def get_state(self):
        return self.bitgen.state

    def set_state(self, state):
        self.bitgen.state = state

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def sample(self, char_exponent, tilt, method=None):
        """
        Generate a random variable from a stable distribution with
            characteristic exponent =  char_exponent < 1
            skewness = 1
            scale = cos(char_exponent * pi / 2) ** (1 / char_exponent)
            location = 0
            exponential tilting = tilt
        (The density p(x) is tilted by exp(- tilt * x).)

        The cost of the divide-conquer algorithm increases as a function of
        'tilt ** char_exp'. While the cost of double-rejection algorithm is
        bounded, the divide-conquer algorithm is simpler and faster for small
        'tilt ** char_exp'.

        References:
        -----------
        Implementation is mostly based on the algorithm descriptions in
            'Sampling Exponentially Tilted Stable Distributions' by Hofert (2011)
        Ideas behind and details on the double-rejection sampling is better
        described in
            'Random variate generation for exponentially and polynomially tilted
            stable distributions' by Devroye (2009)
        """
        if not isinstance(tilt, np.ndarray):
            raise TypeError('Tilt parameter must be a numpy array.')
        if isinstance(char_exponent, (np.floating, float)):
            char_exponent = np.tile(char_exponent, tilt.size)
        elif isinstance(char_exponent, np.ndarray):
            if not char_exponent.size == tilt.size:
                raise ValueError('Input arrays must be of the same length.')
        else:
            raise TypeError('Characteristic exponent must be float or numpy array.')
        if not np.all(char_exponent < 1):
            raise ValueError('Characteristic exponent must be smaller than 1.')
        if not np.all(tilt > 0):
            raise ValueError('Tilting parameter must be positive.')

        if method is None:
            # Choose a likely faster method.
            divide_conquer_cost = tilt ** char_exponent
            double_rejection_cost = self.TILT_POWER_THRESHOLD
                # The relative costs are implementation & architecture dependent.
            use_divide_conquer = (divide_conquer_cost < double_rejection_cost)
        elif method in ['divide-conquer', 'double-rejection']:
            use_divide_conquer = np.tile(method == 'divide-conquer', tilt.size)
        else:
            raise ValueError("Unrecognized method name.")

        char_exponent = char_exponent.astype(np.double)
        tilt = tilt.astype(np.double)
        use_divide_conquer = use_divide_conquer.astype(np.uint8)
        result = np.zeros(tilt.size, dtype=np.double)

        cdef double[:] char_exponent_view = char_exponent
        cdef double[:] tilt_view = tilt
        cdef np_uint8[:] use_divide_conquer_view = use_divide_conquer
        cdef double[:] result_view = result
        cdef long n_sample = tilt.size
        cdef Py_ssize_t i

        for i in range(n_sample):
            if use_divide_conquer_view[i]:
                result_view[i] = self.sample_by_divide_and_conquer(
                    char_exponent_view[i], tilt_view[i]
                )
            else:
                result_view[i] = self.sample_by_double_rejection(
                    char_exponent_view[i], tilt_view[i]
                )
        return result

    cdef double sample_by_divide_and_conquer(self, double char_exp, double tilt):
        cdef double X, c
        cdef long partition_size = max(1, <long>floor(pow(tilt, char_exp)))
        X = 0.
        c = pow(1. / partition_size, 1. / char_exp)
        for i in range(partition_size):
            X += self.sample_divided_rv(char_exp, tilt, c)
        return X

    cdef double sample_divided_rv(self, double char_exp, double tilt, double c):
        cdef bint accepted = False
        while not accepted:
            S = c * self.sample_non_tilted_rv(char_exp)
            accept_prob = exp(- tilt * S)
            accepted = (random_uniform(self.bitgen) < accept_prob)
        return S

    cdef double sample_non_tilted_rv(self, double char_exp):
        cdef double S = pow(
            - self.zolotarev_function(M_PI * random_uniform(self.bitgen), char_exp)
                / log(random_uniform(self.bitgen)),
            (1. - char_exp) / char_exp
        )
        return S

    cdef double sample_by_double_rejection(self, double char_exp, double tilt):

        cdef double U, V, X, z, log_accept_prob
        cdef double tilt_power = pow(tilt, char_exp)

        # Start double-rejection sampling.
        cdef bint accepted = False
        while not accepted:
            U, V, z = self.sample_aux_rv(char_exp, tilt_power)
            X, log_accept_prob = \
                self.sample_reference_rv(U, char_exp, tilt_power, z)
            accepted = (log_accept_prob > log(V))

        return pow(X, - (1. - char_exp) / char_exp)

    cdef (double, double, double) \
        sample_aux_rv(self, double char_exp, double tilt_power):
        """
        Samples an auxiliary random variable for the double-rejection algorithm.
        Returns:
            U : auxiliary random variable for the double-rejection algorithm
            V : uniform random variable independent of U, X
            z : scalar quantity used later
        """
        cdef double U, V, z, accept_prob
        cdef double gamma, xi, psi
            # Intermediate quantities; could be computed outside the funciton
            # and reused in case of rejection
        gamma = tilt_power * char_exp * (1. - char_exp)
        xi = (1. + sqrt(2. * gamma) * (2. + sqrt(.5 * M_PI))) / M_PI
        psi = sqrt(gamma / M_PI) * (2. + sqrt(.5 * M_PI)) \
            * exp(- gamma * M_PI * M_PI / 8.)
        cdef bint accepted = False
        while not accepted:
            U = self.sample_aux2_rv(xi, psi, gamma)
            if U > M_PI:
                continue

            zeta = sqrt(self.zolotarev_pdf_exponentiated(U, char_exp))
            z = 1. / (1. - pow(1. + char_exp * zeta / sqrt(gamma), -1. / char_exp))
            accept_prob = self.compute_aux2_accept_prob(
                U, xi, psi, zeta, z, tilt_power, gamma
            )
            if accept_prob > 0.:
                V = random_uniform(self.bitgen) / accept_prob
                accepted = (U < M_PI and V <= 1.)

        return U, V, z

    cdef double sample_aux2_rv(self,
            double xi, double psi, double gamma):
        """
        Sample the 2nd level auxiliary random variable (i.e. the additional
        auxiliary random variable used to sample the auxilary variable for
        double-rejection algorithm.)
        """

        w1 = sqrt(.5 * M_PI / gamma) * xi
        w2 = 2. * sqrt(M_PI) * psi
        w3 = xi * M_PI
        V = random_uniform(self.bitgen)
        if gamma >= 1:
            if V < w1 / (w1 + w2):
                U = fabs(random_normal(self.bitgen)) / sqrt(gamma)
            else:
                W = random_uniform(self.bitgen)
                U = M_PI * (1. - W * W)
        else:
            W = random_uniform(self.bitgen)
            if V < w3 / (w2 + w3):
                U = M_PI * W
            else:
                U = M_PI * (1. - W * W)

        return U

    cdef double compute_aux2_accept_prob(self,
            double U, double xi, double psi, double zeta, double z,
            double tilt_power, double gamma
        ):
        inverse_accept_prob = M_PI * exp(-tilt_power * (1. - 1. / (zeta * zeta))) \
              / ((1. + sqrt(.5 * M_PI)) * sqrt(gamma) / zeta + z)
        d = 0.
        if U >= 0. and gamma >= 1:
            d += xi * exp(-gamma * U * U / 2.)
        if U > 0. and U < M_PI:
            d += psi / sqrt(M_PI - U)
        if U >= 0. and U <= M_PI and gamma < 1.:
            d += xi
        inverse_accept_prob *= d
        accept_prob = 1 / inverse_accept_prob
        return accept_prob

    cdef (double, double) sample_reference_rv(self,
            double U, double char_exp, double tilt_power, double z):
        """
        Generate a sample from the reference (augmented) distribution conditional
        on U for the double-rejection algorithm. The algorithm use a rejection
        sampler with half-Gaussian, uniform, and truncated exponential to the
        left, middle, and right of a partitioned real-line.

        Returns:
        --------
            X : random variable from the reference distribution
            N, E : random variables used later for computing the acceptance prob
        """
        cdef double a, left_thresh, right_thresh, expo_scale, \
            mass_left, mass_mid, mass_right, mass_total, X, V, N, E
        a = self.zolotarev_function(U, char_exp)
        left_thresh = pow((1. - char_exp) / char_exp / a, char_exp) * tilt_power
        right_thresh = left_thresh + sqrt(left_thresh * char_exp / a)
        expo_scale = z / a
        mass_left = (right_thresh - left_thresh) * sqrt(.5 * M_PI)
        mass_mid = (right_thresh - left_thresh)
        mass_right = expo_scale
        mass_total = mass_left + mass_mid + mass_right
        V = random_uniform(self.bitgen)
        N = 0.
        E = 0.
        # Divided into three pieces at left_thresh and (left_thresh + mid_width)
        if V < mass_left / mass_total:
            N = random_normal(self.bitgen)
            X = left_thresh - (right_thresh - left_thresh) * fabs(N)
        elif V < (mass_left + mass_mid) / mass_total:
            X = left_thresh + (right_thresh - left_thresh) * random_uniform(self.bitgen)
        else:
            E = - log(random_uniform(self.bitgen))
            X = right_thresh + E * mass_right

        log_accept_prob = self.compute_log_accept_prob(
            X, N, E, left_thresh, right_thresh, a, char_exp, tilt_power
        )
        return X, log_accept_prob

    cdef double compute_log_accept_prob(self,
            double X, double N, double E, double left_thresh, double right_thresh,
            double a, double char_exp, double tilt_power
        ):
        cdef double char_exp_odds = (1. - char_exp) / char_exp
        if X < 0:
            log_accept_prob = - INFINITY
        else:
            log_accept_prob = - (
                a * (X - left_thresh)
                + exp(log(tilt_power) / char_exp - char_exp_odds * log(left_thresh))
                * (pow(left_thresh / X, char_exp_odds) - 1.)
            )
            if X < left_thresh:
                log_accept_prob += N * N / 2.
            elif X > right_thresh:
                log_accept_prob += E

        return log_accept_prob

    cdef double zolotarev_pdf_exponentiated(self, double x, double char_exp):
        """
        Evaluates a function proportional to a power of the Zolotarev density.
        """
        cdef double denominator, numerator
        denominator = pow(sinc(char_exp * x), char_exp) \
                      * pow(sinc((1. - char_exp) * x), (1. - char_exp))
        numerator = sinc(x)
        return numerator / denominator

    cdef double zolotarev_function(self, double x, double char_exp):
        cdef double val = pow(
            pow((1. - char_exp) * sinc((1. - char_exp) * x), (1. - char_exp))
            * pow(char_exp * sinc(char_exp * x), char_exp)
            / sinc(x)
        , 1. / (1. - char_exp))
        return val



import math
from warnings import warn
import numpy as np
import scipy as sp
from scipy.special import polygamma as scipy_polygamma

class RegressionCoefPrior():

    def __init__(
            self,
            bridge_exponent=.5,
            n_fixed_effect=0,
            sd_for_intercept=float('inf'),
            sd_for_fixed_effect=float('inf'),
            regularizing_slab_size=float('inf'),
            global_scale_prior_hyper_param=None,
            _global_scale_parametrization='coef_magnitude'
        ):
        """ Encapisulate prior information for BayesBridge.

        Parameters
        ----------
        bridge_exponent : float < 2
            Exponent of the bridge prior on regression coefficients. For example,
            the value of 2 (albeit unsupported) would correspond to Gaussian prior
            and of 1 double-exponential as in Bayesian Lasso.
        n_fixed_effect : int
            Number of predictors --- other than intercept and placed at the
            first columns of the design matrices --- whose coefficients are
            estimated with Gaussian priors of pre-specified standard
            deviation(s).
        sd_for_intercept : float
            Standard deviation of Gaussian prior on the intercept. `Inf`
            corresponds to an uninformative flat prior.
        sd_for_fixed_effect : float, numpy array
            Standard deviation(s) of Gaussian prior(s) on fixed effects.
            If an array, the length must be the same as `n_fixed_effect`.
            `Inf` corresponds to an uninformative flat prior.
        regularizing_slab_size : float
            Standard deviation of the Gaussian tail-regularizer on
            the bridge prior. Used to impose soft prior constraints on a
            range of regression coefficients in case the data provides limited
            information (e.g. when complete separation occurs). One may, for
            example, set the slab size by first choosing a value which
            regression coefficients are very unlikely to exceed in magnitude and
            then dividing the value by 1.96.
        global_scale_prior_hyper_param : dict, None
            Should contain pair of keys 'log10_mean' and 'log10_sd',
            specifying the prior mean and standard deviation of
            log10(global_scale). If None, the default reference prior for a
            scale parameter is used.

        Other Parameters
        ----------------
        _global_scale_parametrization: str, {'raw', 'coef_magnitude'}
            If 'coef_magnitude', scale the local and global scales so that the
            global scale parameter coincide with the prior expected
            magnitude of regression coefficients.
        """
        if not (np.isscalar(sd_for_fixed_effect)
                or n_fixed_effect == len(sd_for_fixed_effect)):
            raise ValueError(
                "Prior sd for fixed effects must be specified either by a "
                "scalar or array of the same length as n_fixed_effect."
            )

        if np.isscalar(sd_for_fixed_effect):
            sd_for_fixed_effect = sd_for_fixed_effect * np.ones(n_fixed_effect)
        self.sd_for_intercept = sd_for_intercept
        self.sd_for_fixed = sd_for_fixed_effect
        self.slab_size = regularizing_slab_size
        self.n_fixed = n_fixed_effect
        self.bridge_exp = bridge_exponent
        self._gscale_paramet = _global_scale_parametrization
        if global_scale_prior_hyper_param is None:
            self.param = {
                'gscale_neg_power': {'shape': 0., 'rate': 0.},
                    # Reference prior for a scale family.
                'gscale': None
            }

        else:
            keys = global_scale_prior_hyper_param.keys()
            if not ({'log10_mean', 'log10_sd'} <= keys):
                raise ValueError(
                    "Dictionary should contain keys 'log10_mean' and 'log10_sd.'"
                )
            log10_mean = global_scale_prior_hyper_param['log10_mean']
            log10_sd = global_scale_prior_hyper_param['log10_sd']
            shape, rate = self.solve_for_gscale_prior_hyperparam(
                log10_mean, log10_sd, bridge_exponent, self._gscale_paramet
            )
            self.param = {
                'gscale_neg_power': {'shape': shape, 'rate': rate},
                'gscale': {'log10_mean': log10_mean, 'log10_sd': log10_sd}
            }   # Hyper-parameters on the negative power are specified in
                # terms of the 'raw' parametrization.

    def get_info(self):
        sd_for_fixed = self.sd_for_fixed
        if len(sd_for_fixed) > 0 and np.all(sd_for_fixed == sd_for_fixed[0]):
            sd_for_fixed = sd_for_fixed[0]
        info = {
            'bridge_exponent': self.bridge_exp,
            'n_fixed_effect': self.n_fixed,
            'sd_for_intercept': self.sd_for_intercept,
            'sd_for_fixed_effect': sd_for_fixed,
            'regularizing_slab_size': self.slab_size,
            'global_scale_prior_hyper_param': self.param['gscale'],
            '_global_scale_parametrization': self._gscale_paramet
        }
        return info

    def clone(self, **kwargs):
        """ Make a clone with only specified attributes modified. """
        info = self.get_info()
        if '_global_scale_parametrization' in kwargs:
            raise ValueError("Change of parametrization is not supported.")
        for key in kwargs.keys():
            if key in info:
                info[key] = kwargs[key]
            else:
                warn("'{:s} is not a valid keyward argument.".format(key))
        return RegressionCoefPrior(**info)

    def adjust_scale(self, gscale, lscale, to):
        unit_bridge_magnitude \
            = self.compute_power_exp_ave_magnitude(self.bridge_exp, 1.)
        if to == 'raw':
            gscale /= unit_bridge_magnitude
            lscale *= unit_bridge_magnitude
        elif to == 'coef_magnitude':
            gscale *= unit_bridge_magnitude
            lscale /= unit_bridge_magnitude
        else:
            raise ValueError()
        return gscale, lscale

    def solve_for_gscale_prior_hyperparam(
            self, log10_mean, log10_sd, bridge_exp, gscale_paramet):
        log_mean = self.change_log_base(log10_mean, from_=10., to=math.e)
        log_sd = self.change_log_base(log10_sd, from_=10., to=math.e)
        if gscale_paramet == 'coef_magnitude':
            unit_bridge_magnitude \
                = self.compute_power_exp_ave_magnitude(bridge_exp, 1.)
            log_mean -= math.log(unit_bridge_magnitude)
        shape, rate = self.solve_for_gamma_param(
            log_mean, log_sd, bridge_exp
        )
        return shape, rate

    @staticmethod
    def compute_power_exp_ave_magnitude(exponent, scale=1.):
        """ Returns the expected absolute value of a random variable with
        density proportional to exp( - |x / scale|^exponent ).
        """
        return scale * math.gamma(2 / exponent) / math.gamma(1 / exponent)

    @staticmethod
    def change_log_base(val, from_=math.e, to=10.):
        return val * math.log(from_) / math.log(to)

    def solve_for_gamma_param(self, log_mean, log_sd, bridge_exp):
        """ Find hyper-parameters matching specified mean and sd in log scale.

        Determine the shape and rate parameters of a Gamma prior on
            phi = gscale ** (- 1 / bridge_exp)
        so that the mean and sd of log(phi) coincide with log_mean and log_sd.
        The calculations are done in the 'raw' parametrization of gscale,
        as opposed to the 'coef_magnitude' parametrization.
        """

        f = lambda log_shape: (
            math.sqrt(self._polygamma(1, math.exp(log_shape))) / bridge_exp
            - log_sd
        ) # Function whose root coincides with the desired log-shape parameter.
        lower_lim = -10.  # Any sufficiently small number is fine.
        if log_sd < 0:
            raise ValueError("Variance has to be positive.")
        elif log_sd > 10 ** 8:
            raise ValueError("Specified prior variance is too large.")
        lower, upper = self._find_root_bounds(f, lower_lim)

        try:
            log_shape = sp.optimize.brentq(f, lower, upper)
        except BaseException as error:
            print('Solving for the global scale gamma prior hyper-parameters '
                  'failed; {}'.format(error))
        shape = math.exp(log_shape)
        rate = math.exp(
            self._polygamma(0, shape) + bridge_exp * log_mean
        )
        return shape, rate

    @staticmethod
    def _polygamma(n, x):
        """ Wrap the scipy function so that it returns a scalar. """
        return scipy_polygamma([n], x)[0]

    @staticmethod
    def _find_root_bounds(f, init_lower_lim, increment=5., max_lim=None):
        if max_lim is None:
            max_lim = init_lower_lim + 10 ** 4
        if f(init_lower_lim) < 0:
            raise ValueError(
                "Objective function must have positive value "
                "at the lower limit."
            )
        lower_lim = init_lower_lim
        while f(lower_lim + increment) > 0 and lower_lim < max_lim:
            lower_lim += increment
        if lower_lim >= max_lim:
            raise Exception()  # Replace with a warning.
        upper_lim = lower_lim + increment
        return (lower_lim, upper_lim)


import abc
import numpy as np
import scipy as sp
import scipy.sparse
import warnings

class AbstractDesignMatrix():

    def __init__(self):
        self.dot_count = 0
        self.Tdot_count = 0
        self.memoized = False
        self.X_dot_v = None # For memoization
        self.v_prev = None # For memoization

    @property
    @abc.abstractmethod
    def shape(self):
        pass

    @abc.abstractmethod
    def dot(self, v):
        pass

    @abc.abstractmethod
    def Tdot(self, v):
        """ Multiply by the transpose of the matrix. """
        pass

    @property
    @abc.abstractmethod
    def is_sparse(self):
        pass

    def memoize_dot(self, flag=True):
        self.memoized = flag
        if self.v_prev is None:
            self.v_prev = np.full(self.shape[1], float('nan'))
        if not flag:
            self.X_dot_v = None
            self.v_prev = None

    @abc.abstractmethod
    def compute_fisher_info(self, weight, diag_only):
        """ Computes X' diag(weight) X and returns it as a numpy array. """
        pass

    @property
    def n_matvec(self):
        return self.dot_count + self.Tdot_count

    def get_dot_count(self):
        return self.dot_count, self.Tdot_count

    def reset_matvec_count(self, count=0):
        if not hasattr(count, "__len__"):
            count = 2 * [count]
        self.dot_count = count[0]
        self.Tdot_count = count[1]

    @abc.abstractmethod
    def toarray(self):
        """ Returns a 2-dimensional numpy array. """
        pass

    @staticmethod
    def remove_intercept_indicator(X):
        if sp.sparse.issparse(X):
            col_variance = np.squeeze(np.array(
                X.power(2).mean(axis=0) - np.power(X.mean(axis=0), 2)
            ))
        else:
            col_variance = np.var(X, axis=0)
        has_zero_variance = (col_variance < X.shape[0] * 2 ** -52)
        if np.any(has_zero_variance):
            warnings.warn(
                "Intercept column (or numerically indistinguishable from "
                "such) detected. Do not add intercept manually. Removing...."
            )
            X = X[:, np.logical_not(has_zero_variance)]
        return X